# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['ImportImage']

from PIL import Image

import TermTk as ttk

class TTkImageNew(ttk.TTkWidget):
    FULLBLOCK = 0x00
    HALFBLOCK = 0x01
    QUADBLOCK = 0x02
    SEXBLOCK   = 0x03

    _quadMap =[
        # 0x00 0x01 0x02 0x03
          ' ', '▘', '▝', '▀',
        # 0x04 0x05 0x06 0x07
          '▖', '▌', '▞', '▛',
        # 0x08 0x09 0x0A 0x0B
          '▗', '▚', '▐', '▜',
        # 0x0C 0x0D 0x0E 0x0F
          '▄', '▙', '▟', '█']

    _sexMap  = [
        # 0x00 0x01 0x02 0x03 0x04 0x05 0x06 0x07
          ' ', '🬀', '🬁', '🬂', '🬃', '🬄', '🬅', '🬆',
        # 0x08 0x09 0x0A 0x0B 0x0C 0x0D 0x0E 0x0F
          '🬇', '🬈', '🬉', '🬊', '🬋', '🬌', '🬍', '🬎',

        # 0x10 0x11 0x12 0x13 0x14 0x15 0x16 0x17
          '🬏', '🬐', '🬑', '🬒', '🬓', '▌', '🬔', '🬕',
        # 0x18 0x19 0x1A 0x1B 0x1C 0x1D 0x1E 0x1F
          '🬖', '🬗', '🬘', '🬙', '🬚', '🬛', '🬜', '🬝',

        # 0x20 0x21 0x22 0x23 0x24 0x25 0x26 0x27
          '🬞', '🬟', '🬠', '🬡', '🬢', '🬣', '🬤', '🬥',
        # 0x28 0x29 0x2A 0x2B 0x2C 0x2D 0x2E 0x2F
          '🬦', '🬧', '▐', '🬨', '🬩', '🬪', '🬫', '🬬',

        # 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37
          '🬭', '🬮', '🬯', '🬰', '🬱', '🬲', '🬳', '🬴',
        # 0x38 0x39 0x3A 0x3B 0x3C 0x3D 0x3E 0x3F
          '🬵', '🬶', '🬷', '🬸', '🬹', '🬺', '🬻', '█']

    __slots__ = ('_data', '_rasterType', '_canvasImage', '_alphaThreshold')
    def __init__(self, **kwargs):
        self._alphaThreshold = 127
        self._rasterType = kwargs.get('rasteriser' , ttk.TTkImage.QUADBLOCK )
        self._data = kwargs.get('data' , [] )
        self._canvasImage = ttk.TTkCanvas()
        self._canvasImage.setTransparent(True)
        super().__init__(**kwargs)
        if self._data:
            self.setData(self._data)

    def toTTkString(self):
        return ttk.TTkString(self._canvasImage.toAnsi())

    def alphaThrteshold(self):
        return self._alphaThreshold

    @ttk.pyTTkSlot(int)
    def setAlphaThreshold(self, value):
        value = max(0,min(255,value))
        if value == self._alphaThreshold:return
        self._alphaThreshold = value
        self._drawImage(self._canvasImage)
        self.update()

    def setData(self, data):
        self._data = data
        w = min(len(i) for i in self._data)
        h = len(self._data)
        if w>0<h and len(data[0][0])==3: # Add alpha channel if missing
            self._data = [[(r,g,b,255) for r,g,b in row] for row in data]
        if self._rasterType == ttk.TTkImage.FULLBLOCK:
            w,h = w,h
        elif self._rasterType == ttk.TTkImage.HALFBLOCK:
            w,h = w,h//2
        elif self._rasterType == ttk.TTkImage.QUADBLOCK:
            w,h = w//2,h//2
        elif self._rasterType == ttk.TTkImage.SEXBLOCK:
            w,h = w//2,h//3
        self._canvasImage.resize(*(self.size()))
        self.resize(w,h)
        self._canvasImage.resize(w,h)
        self._canvasImage.updateSize()
        self._drawImage(self._canvasImage)
        self.update()

    def setRasteriser(self, rasteriser):
        if self._rasterType == rasteriser: return
        self._rasterType = rasteriser
        if self._data:
            self.setData(self._data)

    def _reduceQuad(self, a,b,c,d):
        # quadblitter notcurses like
        th = self._alphaThreshold
        l = [a,b,c,d]
        lth = [(px,1<<i) for i,px in enumerate(l) if px[3]>=th]
        if not lth:
            # Transparent block
            return ttk.TTkString(' ')

        if (ll:=len(lth))<4:
            # Semi-Transparent Block
            cr = sum(px[0] for px,_ in lth)//ll
            cg = sum(px[1] for px,_ in lth)//ll
            cb = sum(px[2] for px,_ in lth)//ll
            ch = sum(g     for  _,g in lth)
            color = ttk.TTkColor.fg(f'#{cr:02X}{cg:02X}{cb:02X}')
            return ttk.TTkString(ttk.TTkImage._quadMap[ch],color)

        if a[:3]==b[:3]==c[:3]==d[:3]:
            color = ttk.TTkColor.bg(f'#{a[0]:02X}{a[1]:02X}{a[2]:02X}')
            return ttk.TTkString(' ',color)

        def delta(i):
            return max(v[i] for v in l) - min(v[i] for v in l)
        deltaR = delta(0)
        deltaG = delta(1)
        deltaB = delta(2)

        def midColor(c1,c2):
            return ((c1[0]+c2[0])//2,(c1[1]+c2[1])//2,(c1[2]+c2[2])//2)

        def closer(a,b,c):
            return \
                ( (a[0]-c[0])**2 + (a[1]-c[1])**2 + (a[2]-c[2])**2 ) > \
                ( (b[0]-c[0])**2 + (b[1]-c[1])**2 + (b[2]-c[2])**2 )

        def splitReduce(i):
            s = sorted(l,key=lambda x:x[i])
            mid = (s[3][i]+s[0][i])//2
            if s[1][i] < mid:
                if s[2][i] > mid:
                    c1 = midColor(s[0],s[1])
                    c2 = midColor(s[2],s[3])
                else:
                    c1 = midColor(s[0],s[1])
                    c1 = midColor(c1,s[2])
                    c2 = s[3]
            else:
                c1 = s[0]
                c2 = midColor(s[1],s[2])
                c2 = midColor(c1,s[3])


            ch  = 0x01 if closer(c1,c2,l[0]) else 0
            ch |= 0x02 if closer(c1,c2,l[1]) else 0
            ch |= 0x04 if closer(c1,c2,l[2]) else 0
            ch |= 0x08 if closer(c1,c2,l[3]) else 0

            return  ttk.TTkString() + \
                    (ttk.TTkColor.bg(f'#{c1[0]:02X}{c1[1]:02X}{c1[2]:02X}') +
                     ttk.TTkColor.fg(f'#{c2[0]:02X}{c2[1]:02X}{c2[2]:02X}')) + \
                    ttk.TTkImage._quadMap[ch]

        if deltaR >= deltaG and deltaR >= deltaB:
            # Use Red as splitter
            return splitReduce(0)
        elif deltaG >= deltaB and deltaG >= deltaR:
            # Use Green as splitter
            return splitReduce(1)
        else:
            # Use Blue as splitter
            return splitReduce(2)

    def _reduceSex(self, a,b,c,d,e,f):
        # quadblitter notcurses like
        th = self._alphaThreshold
        l = [a,b,c,d,e,f]
        lth = [(px,1<<i) for i,px in enumerate(l) if px[3]>=th]
        if not lth:
            # Transparent block
            return ttk.TTkString(' ')

        if (ll:=len(lth))<4:
            # Semi-Transparent Block
            cr = sum(px[0] for px,_ in lth)//ll
            cg = sum(px[1] for px,_ in lth)//ll
            cb = sum(px[2] for px,_ in lth)//ll
            ch = sum(g     for  _,g in lth)
            color = ttk.TTkColor.fg(f'#{cr:02X}{cg:02X}{cb:02X}')
            return ttk.TTkString(ttk.TTkImage._sexMap[ch],color)

        if a[:3]==b[:3]==c[:3]==d[:3]==e[:3]==f[:3]:
            color = ttk.TTkColor.bg(f'#{a[0]:02X}{a[1]:02X}{a[2]:02X}')
            return ttk.TTkString(' ',color)

        def delta(i):
            return max(v[i] for v in l) - min(v[i] for v in l)
        deltaR = delta(0)
        deltaG = delta(1)
        deltaB = delta(2)

        def midColor(c1,c2):
            return ((c1[0]+c2[0])//2,(c1[1]+c2[1])//2,(c1[2]+c2[2])//2)

        def closer(a,b,c):
            return \
                ( (a[0]-c[0])**2 + (a[1]-c[1])**2 + (a[2]-c[2])**2 ) > \
                ( (b[0]-c[0])**2 + (b[1]-c[1])**2 + (b[2]-c[2])**2 )

        def splitReduce(i):
            s = sorted(l,key=lambda x:x[i])
            mid = (s[5][i]+s[0][i])//2
            if s[1][i] < mid:
                if s[2][i] > mid:
                    c1 = midColor(s[0],s[1])
                    c2 = midColor(s[2],s[3])
                else:
                    c1 = midColor(s[0],s[1])
                    c1 = midColor(c1,s[2])
                    c2 = s[3]
            else:
                c1 = s[0]
                c2 = midColor(s[1],s[2])
                c2 = midColor(c1,s[3])


            ch  = 0x01 if closer(c1,c2,l[0]) else 0
            ch |= 0x02 if closer(c1,c2,l[1]) else 0
            ch |= 0x04 if closer(c1,c2,l[2]) else 0
            ch |= 0x08 if closer(c1,c2,l[3]) else 0
            ch |= 0x10 if closer(c1,c2,l[4]) else 0
            ch |= 0x20 if closer(c1,c2,l[5]) else 0

            return  ttk.TTkString() + \
                    (ttk.TTkColor.bg(f'#{c1[0]:02X}{c1[1]:02X}{c1[2]:02X}') +
                     ttk.TTkColor.fg(f'#{c2[0]:02X}{c2[1]:02X}{c2[2]:02X}')) + \
                    ttk.TTkImage._sexMap[ch]

        if deltaR >= deltaG and deltaR >= deltaB:
            # Use Red as splitter
            return splitReduce(0)
        elif deltaG >= deltaB and deltaG >= deltaR:
            # Use Green as splitter
            return splitReduce(1)
        else:
            # Use Blue as splitter
            return splitReduce(2)

    def rotHue(self, deg):
        old = self._data
        self._data = [[p for p in l ] for l in old]
        for row in self._data:
            for i,pixel in enumerate(row):
                h,s,l = ttk.TTkColor.rgb2hsl(pixel)
                row[i] = ttk.TTkColor.hsl2rgb(((h+deg)%360,s,l))
        self.setData(self._data)

    def _drawImage(self, canvas):
        img = self._data
        threshold = self._alphaThreshold
        if self._rasterType == ttk.TTkImage.FULLBLOCK:
            for y in range(0, len(img)):
                for x in range(0, len(img[y])):
                    r,g,b,a = img[y][x]
                    color = ttk.TTkColor.bg(f'#{r:02X}{g:02X}{b:02X}') if a>threshold else ttk.TTkColor.RST
                    canvas.drawChar(pos=(x,y), char=' ', color=color)
        elif self._rasterType == ttk.TTkImage.HALFBLOCK:
            for y in range(0, len(img)&(~1), 2):
                for x in range(0, len(img[y])):
                    c1, c2 = img[y][x] ,img[y+1][x]
                    if c1[3]<threshold and c2[3]<threshold:
                        canvas.drawChar(pos=(x,y//2), char=' ', color=ttk.TTkColor.RST)
                    elif c2[3]<threshold:
                        color = ttk.TTkColor.fg(f'#{c1[0]:02X}{c1[1]:02X}{c1[2]:02X}')
                        canvas.drawChar(pos=(x,y//2), char='▀', color=color)
                    elif c1[3]<threshold:
                        color = ttk.TTkColor.fg(f'#{c2[0]:02X}{c2[1]:02X}{c2[2]:02X}')
                        canvas.drawChar(pos=(x,y//2), char='▄', color=color)
                    elif c1[:3]==c2[:3]:
                        color = ttk.TTkColor.bg(f'#{c1[0]:02X}{c1[1]:02X}{c1[2]:02X}')
                        canvas.drawChar(pos=(x,y//2), char=' ', color=color)
                    else:
                        color = ( ttk.TTkColor.fg(f'#{c1[0]:02X}{c1[1]:02X}{c1[2]:02X}') +
                                  ttk.TTkColor.bg(f'#{c2[0]:02X}{c2[1]:02X}{c2[2]:02X}') )
                        canvas.drawChar(pos=(x,y//2), char='▀', color=color)
        elif self._rasterType == ttk.TTkImage.QUADBLOCK:
            for y in range(0, len(img)&(~1), 2):
                for x in range(0, min(len(img[y])&(~1),len(img[y+1])&(~1)), 2):
                    canvas.drawText(
                            pos=(x//2,y//2),
                            text=self._reduceQuad(
                                        img[y][x]   , img[y][x+1]   ,
                                        img[y+1][x] , img[y+1][x+1] ))
        elif self._rasterType == ttk.TTkImage.SEXBLOCK:
            for y in range(0, len(img)-2, 3):
                for x in range(0, min(len(img[y])-1,len(img[y+1])-1,len(img[y+2])-1), 2):
                    canvas.drawText(
                            pos=(x//2,y//3),
                            text=self._reduceSex(
                                        img[y][x]   , img[y][x+1]   ,
                                        img[y+1][x] , img[y+1][x+1] ,
                                        img[y+2][x] , img[y+2][x+1] ))

    def paintEvent(self, canvas: ttk.TTkCanvas):
        w,h=self.size()
        s = (0,0,w,h)
        for y,(rowd,rowc) in enumerate(zip(self._canvasImage._data[:h],self._canvasImage._colors[:h])):
            for x,(gl,c) in enumerate(zip(rowd[:w],rowc[:w])):
                nfg = c.foreground()
                nbg = c.background()
                if gl == ' ' and not nbg:
                    continue
                elif gl == ' ':
                    canvas._data[  y][x] = ' '
                    canvas._colors[y][x] = nbg
                elif not nbg:
                    nbg = canvas._colors[y][x].background()
                    canvas._data[  y][x] = gl
                    canvas._colors[y][x] = nbg + nfg if nbg else nfg
                else:
                    canvas._data[  y][x] = gl
                    canvas._colors[y][x] = c
        # canvas.paintCanvas(self._canvasImage,s,s,s)

class ImagePreview(TTkImageNew):
    __slots__ = ('_trColor1','_trColor2')
    def __init__(self, **kwargs):
        self._trColor1 = ttk.TTkColor.bg("#777777")
        self._trColor2 = ttk.TTkColor.bg("#bbbbbb")
        super().__init__(**kwargs)

    def setTransparentColor(self, c:ttk.TTkColor):
        r,g,b = c.bgToRGB()
        if r+g+b < 127*3:
            r1,g1,b1 = 200,200,200
            r2,g2,b2 = 220,220,220
        else:
            r1,g1,b1 =  80, 80, 80
            r2,g2,b2 = 100,100,100
        self._trColor1 = ttk.TTkColor.bg(f'#{r1:02X}{g1:02X}{b1:02X}')
        self._trColor2 = ttk.TTkColor.bg(f'#{r2:02X}{g2:02X}{b2:02X}')
        self.update()

    def paintEvent(self, canvas: ttk.TTkCanvas):
        w,h = self.size()
        ws,hs=8,4
        c = [self._trColor1,self._trColor2]
        for y in range(1+h//hs):
            for x in range(1+w//ws):
                canvas.fill(pos=(x*ws,y*hs),size=(ws,hs),color=c[(x+y)%2])
        super().paintEvent(canvas)

class ImportImage(ttk.TTkWindow):
    __slots__ = ('_pilImage_bk','_pilImage','_image','_b_width','_b_height','_cb_resample','_b_color',
                 'exportedImage')
    def __init__(self, pilImage:Image, **kwargs):
        self.exportedImage = ttk.pyTTkSignal(ttk.TTkString)
        self._pilImage =  self._pilImage_bk = pilImage.convert('RGBA')
        layout = ttk.TTkGridLayout()
        super().__init__(**kwargs|{"layout":layout,'size':(100,30)})

        saImage = ttk.TTkScrollArea()
        self._image = image = ImagePreview(parent=saImage.viewport())
        image.setRasteriser(ttk.TTkImage.HALFBLOCK)

        resizeFrame     = ttk.TTkFrame(title='Resize',layout=ttk.TTkGridLayout())
        propertiesFrame = ttk.TTkFrame(title='Image Properties',layout=ttk.TTkGridLayout())

        resizeFrame.layout().addWidget(ttk.TTkLabel(text='Width:'   ),0,0)
        resizeFrame.layout().addWidget(b_width  := ttk.TTkSpinBox(maximum=0x1000),0,1)
        resizeFrame.layout().addWidget(ttk.TTkLabel(text='Height:'  ),1,0)
        resizeFrame.layout().addWidget(b_height := ttk.TTkSpinBox(maximum=0x1000),1,1)
        resizeFrame.layout().addWidget(ttk.TTkLabel(text='Resample:'),2,0)
        resizeFrame.layout().addWidget(cb_resample := ttk.TTkComboBox(),2,1)

        resizeFrame.layout().addWidget(b_quadAR := ttk.TTkButton(text='Quad A/R', border=True),0,2,3,1)
        resizeFrame.layout().addWidget(b_sexAR  := ttk.TTkButton(text='Sex A/R', border=True), 0,3,3,1)
        resizeFrame.layout().addWidget(b_resize := ttk.TTkButton(text='Resize', border=True),  0,4,3,1)

        resizeFrame.layout().addWidget(ttk.TTkLabel(text='Adjust:'),3,0)
        sl_res = ttk.TTkSlider(value=100, minimum=1, maximum=200, orientation=ttk.TTkK.HORIZONTAL)
        resizeFrame.layout().addWidget(sl_res, 3,1,1,4)

        cb_resample.addItems(['NEAREST','BOX','BILINEAR','HAMMING','BICUBIC','LANCZOS'])
        cb_resample.setCurrentIndex(0)

        propertiesFrame.layout().addWidget(ttk.TTkLabel(text='Resolution:',maxWidth=11), 0,0)
        propertiesFrame.layout().addWidget(cb_resolution := ttk.TTkComboBox(),           0,1,1,2)
        propertiesFrame.layout().addWidget(ttk.TTkLabel(text='AlphaColor:'),             1,0)
        propertiesFrame.layout().addWidget(b_color  := ttk.TTkColorButtonPicker(color=ttk.TTkColor.fg("#000000")),  1,1,1,2)
        propertiesFrame.layout().addWidget(b_export := ttk.TTkButton(text="Export"),     3,0,1,3)
        # propertiesFrame.layout().addItem(ttk.TTkLayout(),3,0,1,2)

        propertiesFrame.layout().addWidget(ttk.TTkLabel(text='AlphaTres.:'),             2,0)
        sl_alpha_tre = ttk.TTkSlider(value=127, minimum=0, maximum=255, orientation=ttk.TTkK.HORIZONTAL)
        propertiesFrame.layout().addWidget(sl_alpha_tre, 2,1,1,2)

        cb_resolution.addItems(['FULLBLOCK','HALFBLOCK','QUADBLOCK','SEXBLOCK'])
        cb_resolution.setCurrentIndex(1)

        layout.addWidget(saImage        ,0,0,1,2)
        layout.addWidget(resizeFrame    ,1,0)
        layout.addWidget(propertiesFrame,1,1)

        self._b_width     = b_width
        self._b_height    = b_height
        self._b_color     = b_color
        self._cb_resample = cb_resample

        width, height = pilImage.size
        b_width.setValue(width)
        b_height.setValue(height)

        @ttk.pyTTkSlot()
        def _quadAR():
            if not self._pilImage: return
            width, height = self._pilImage.size
            b_width.setValue(width)
            b_height.setValue(height//2)

        @ttk.pyTTkSlot()
        def _sexAR():
            if not self._pilImage: return
            width, height = self._pilImage.size
            # w/h = 4/3
            b_width.setValue(width)
            b_height.setValue(height*3//4)

        @ttk.pyTTkSlot(int)
        def _adjustResolution(val):
            if not self._pilImage: return
            width, height = self._pilImage.size
            b_width.setValue(val*width//100)
            b_height.setValue(val*height//100)

        @ttk.pyTTkSlot(str)
        def _resolutionChanged(res):
            newRes = {
                'FULLBLOCK' : ttk.TTkImage.FULLBLOCK,
                'HALFBLOCK' : ttk.TTkImage.HALFBLOCK,
                'QUADBLOCK' : ttk.TTkImage.QUADBLOCK,
                'SEXBLOCK'  : ttk.TTkImage.SEXBLOCK
                    }.get(res, ttk.TTkImage.QUADBLOCK)
            image.setRasteriser(newRes)

        b_export.clicked.connect(self.export)
        b_quadAR.clicked.connect(_quadAR)
        b_sexAR.clicked.connect(_sexAR)
        b_resize.clicked.connect(self._resize)
        b_color.colorSelected.connect(self._updateImage)
        cb_resolution.currentTextChanged.connect(_resolutionChanged)
        sl_alpha_tre.valueChanged.connect(image.setAlphaThreshold)
        sl_res.valueChanged.connect(_adjustResolution)

        self._updateImage()

    @ttk.pyTTkSlot()
    def _resize(self):
        if not (pilImage := self._pilImage): return
        w = self._b_width.value()
        h = self._b_height.value()
        resample = {'NEAREST' : Image.NEAREST,
                    'BOX' :     Image.BOX,
                    'BILINEAR': Image.BILINEAR,
                    'HAMMING' : Image.HAMMING,
                    'BICUBIC' : Image.BICUBIC,
                    'LANCZOS' : Image.LANCZOS}.get(
                        self._cb_resample.currentText(),Image.NEAREST)
        self._pilImage = self._pilImage_bk.resize((w,h),resample)
        self._updateImage()

    @ttk.pyTTkSlot()
    def _updateImage(self):
        data = list(self._pilImage.getdata())
        w,h = self._pilImage.size
        br,bg,bb = self._b_color.color().bgToRGB()
        rgbList = [
            ((r*a+(255-a)*br)//255,
            (g*a+(255-a)*bg)//255,
            (b*a+(255-a)*bb)//255, a)
            for r,g,b,a in data]

        self._image.setTransparentColor(self._b_color.color())
        imageList = [rgbList[i:i+w] for i in range(0, len(rgbList), w)]
        self._image.setData(imageList)

    @ttk.pyTTkSlot()
    def export(self):
        self.exportedImage.emit(self._image.toTTkString())



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['Brush']

import TermTk as ttk

from .const import ToolType

# Due bto the variety of widgets that depends
# on the brush state, I am defining this class
# with all the signals required to broadcast
# all the changes to the widgets that require it
class Brush():
    __slots__ = ['_color','_glyph','_area','_toolType',
                 '_fgEnabled','_bgEnabled','_glyphEnabled',
                 '_transparentArea',
                 #signals
                 'colorChanged', 'glyphChanged', 'areaChanged',
                 'toolTypeChanged',
                 'fgEnabledChanged','bgEnabledChanged','glyphEnabledChanged'
                 ]
    def __init__(self) -> None:
        self.fgEnabledChanged    = ttk.pyTTkSignal(bool)
        self.bgEnabledChanged    = ttk.pyTTkSignal(bool)
        self.glyphEnabledChanged = ttk.pyTTkSignal(bool)
        self.areaChanged      = ttk.pyTTkSignal(ttk.TTkString)
        self.colorChanged     = ttk.pyTTkSignal(ttk.TTkColor)
        self.glyphChanged     = ttk.pyTTkSignal(str)
        self.toolTypeChanged  = ttk.pyTTkSignal(ToolType)

        self._area  : ttk.TTkString = ttk.TTkString()
        self._color : ttk.TTkColor  = ttk.TTkColor.RST
        self._glyph : str = 'X'
        self._fgEnabled:     bool= True
        self._bgEnabled:     bool= True
        self._glyphEnabled:  bool= True
        self._toolType: ToolType = ToolType.BRUSH | ToolType.GLYPH

    def toolType(self) -> ToolType:
        return self._toolType

    def setToolType(self, ty:ToolType) -> None:
        if self._toolType == ty: return
        self._toolType = ty
        self.toolTypeChanged.emit(ty)

    def color(self) -> ttk.TTkColor:
        return self._color

    def setColor(self, color:ttk.TTkColor) -> None:
        if self._color == color: return
        self._color = color
        self.colorChanged.emit(color)

    def glyph(self) -> str:
        return self._glyph

    def setGlyph(self, glyph:str) -> None:
        if self._glyph == glyph: return
        self._glyph = glyph
        self.glyphChanged.emit(glyph)

    def area(self) -> ttk.TTkString:
        return self._area

    def setArea(self, area:ttk.TTkString) -> None:
        if self._area == area: return
        self._area = area
        self.areaChanged.emit(area)

    def glyphEnabled(self) -> bool:
        return self._glyphEnabled

    def setFgEnabled(self, enabled:bool) -> None:
        if self._fgEnabled == enabled: return
        self._fgEnabled = enabled
        self.fgEnabledChanged.emit(enabled)

    def fgEnabled(self) -> bool:
        return self._fgEnabled

    def setBgEnabled(self, enabled:bool) -> None:
        if self._bgEnabled == enabled: return
        self._bgEnabled = enabled
        self.bgEnabledChanged.emit(enabled)

    def bgEnabled(self) -> bool:
        return self._bgEnabled

    def setGlyphEnabled(self, enabled:bool) -> None:
        if self._glyphEnabled == enabled: return
        self._glyphEnabled = enabled
        self.glyphEnabledChanged.emit(enabled)



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['PaintArea','PaintScrollArea','CanvasLayer']

import TermTk as ttk

from .canvaslayer import CanvasLayer
from .const       import ToolType
from .glbls       import glbls

class PaintArea(ttk.TTkAbstractScrollView):
    __slots__ = ('_canvasLayers', '_currentLayer',
                 '_transparentColor',
                 '_documentPos','_documentSize',
                 '_mouseMove', '_mouseDrag', '_mousePress', '_mouseRelease',
                 '_moveData','_resizeData',
                 '_clipboard',
                 '_tool',
                 '_glyph', '_glyphColor', '_glyphEnabled', '_areaBrush',
                 # Signals
                 'layerSelected', 'layerAdded')

    def __init__(self, *args, **kwargs):
        self.layerSelected = ttk.pyTTkSignal(CanvasLayer)
        self.layerAdded    = ttk.pyTTkSignal(CanvasLayer)
        self._transparentColor = {'base':ttk.TTkColor.RST,'dim':ttk.TTkColor.RST}
        self._currentLayer:CanvasLayer = None
        self._canvasLayers:list[CanvasLayer] = []
        self._glyph = 'X'
        self._glyphColor = ttk.TTkColor.RST
        self._glyphEnabled = True
        self._areaBrush = CanvasLayer()
        self._areaBrush.changed.connect(self.update)
        self._moveData = None
        self._resizeData = None
        self._mouseMove = None
        self._mouseDrag = None
        self._mousePress   = None
        self._mouseRelease = None
        self._tool = 0
        self._documentPos    = (6,3)
        self._documentSize   = ( 0, 0)
        self._clipboard = ttk.TTkClipboard()
        super().__init__(*args, **kwargs)
        self.setTrans(ttk.TTkColor.bg('#FF00FF'))
        self.resizeCanvas(80,25)
        self.setFocusPolicy(ttk.TTkK.ClickFocus + ttk.TTkK.TabFocus)

        glbls.brush.toolTypeChanged.connect(self.setTool)
        glbls.brush.areaChanged.connect(    self.setAreaBrush)
        glbls.brush.glyphChanged.connect(       self.updateGlyph)
        glbls.brush.colorChanged.connect(       self.updateGlyph)
        glbls.brush.glyphEnabledChanged.connect(self.updateGlyph)

        # Retrieve the default values
        self.setTool(      glbls.brush.toolType())
        self.updateGlyph()

    def clear(self):
        self._documentPos    = (6,3)
        self._documentSize   = (80, 25)
        self._currentLayer:CanvasLayer = None
        self._canvasLayers:list[CanvasLayer] = []
        self.update()

    def _getGeometry(self):
        dx,dy = self._documentPos
        dw,dh = self._documentSize
        ww,wh = self.size()
        x1,y1 = min(0,dx),min(0,dy)
        x2,y2 = max(dx+dw,ww),max(dy+dh,wh)
        for l in self._canvasLayers:
            lx,ly = l.pos()
            lw,lh = l.size()
            x1 = min(x1,dx+lx)
            y1 = min(y1,dy+ly)
            x2 = max(x2,dx+lx+lw)
            y2 = max(y2,dy+ly+lh)
        # ttk.TTkLog.debug(f"{x1=},{y1=},{x2-x1=},{y2-y1=}")
        return x1,y1,x2-x1,y2-y1

    def _retuneGeometry(self):
            ox, oy = self.getViewOffsets()
            dx,dy = self._documentPos
            x1,y1,_,_ = self._getGeometry()
            dx1,dy1 = max(0,dx,dx-x1),max(0,dy,dy-y1)
            self._documentPos = dx1,dy1
            self.viewChanged.emit()
            # dx,dy = self._documentPos
            # self.chan
            self.viewMoveTo(ox+dx1-dx,oy+dy1-dy)
            # If the area move to be adapted to the
            # Negative coordinates, the reference values used in
            # mouse press, moveData, resizeData need to be
            # adapted to the new topology
            if mp:=self._mousePress:
                mpx,mpy = mp
                self._mousePress = (mpx-x1,mpy-y1)
            if md:=self._moveData:
                mx,my=md['pos']
                md['pos']=(mx+dx1-dx,my+dy1-dy)
            # if rd:=self._resizeData:
            #     rx,ry,rw,rh=rd['geometry']
            #     # rd['geometry']=(rx+dx1-dx,ry+dy1-dy,rw,rh)

    def viewFullAreaSize(self) -> tuple[int,int]:
        _,_,w,h = self._getGeometry()
        return w,h

    def viewDisplayedSize(self) -> tuple:
        return self.size()

    def maximumWidth(self):   return 0x10000
    def maximumHeight(self):  return 0x10000
    def minimumWidth(self):   return 0
    def minimumHeight(self):  return 0

    def canvasLayers(self):
        return self._canvasLayers

    def resizeCanvas(self, w, h):
        self._documentSize  = (w,h)
        self._retuneGeometry()
        self.update()

    def superResize(self,x,y,w,h):
        dx,dy = self._documentPos
        dw,dh = self._documentSize
        if (x,y,w,h) == (dx,dy,dw,dh): return
        if w<0: x=dx;w=dw
        if h<0: y=dy;h=dh
        if self._documentPos  != (x,y):
            diffx = dx-x
            diffy = dy-y
            for l in self._canvasLayers:
                lx,ly = l.pos()
                l.move(lx+diffx,ly+diffy)
            self._documentPos = (x,y)
        self._documentSize = (w,h)
        self.update()

    def setCurrentLayer(self, layer:CanvasLayer):
        self._currentLayer = layer
        self.update()

    def newLayer(self) -> CanvasLayer:
        newLayer = CanvasLayer()
        w,h = self._documentSize
        w,h = (w,h) if (w,h)!=(0,0) else (80,24)
        newLayer.resize(w,h)
        self._currentLayer = newLayer
        newLayer.changed.connect(self.update)
        self._canvasLayers.append(newLayer)
        self.layerAdded.emit(newLayer)
        self._retuneGeometry()
        return newLayer

    def importLayer(self, dd):
        self._currentLayer.importLayer(dd)
        self._retuneGeometry()

    def importDocument(self, dd):
        self._canvasLayers = []
        if (
            ( 'version' in dd and dd['version'] == '1.0.0' ) or
            ( 'version' in dd and dd['version'] == '1.0.1' and dd['type'] == 'DumbPaintTool/Document') ):
            self.resizeCanvas(*dd['size'])
            for l in dd['layers']:
                nl = self.newLayer()
                nl.importLayer(l)
        else:
            ttk.TTkLog.error("File Format not recognised")
        self._retuneGeometry()

    def exportLayer(self, full=False, palette=True, crop=True) -> dict:
        if self._currentLayer:
            return self._currentLayer.exportLayer(full=full,palette=palette,crop=crop)
        return {}

    def exportDocument(self, full=True, palette=True, crop=True) -> dict:
        pw,ph = self._documentSize
        outData  = {
            'type':'DumbPaintTool/Document',
            'version':'1.0.1',
            'size':(pw,ph),
            'layers':[l.exportLayer(full=full,palette=palette,crop=crop) for l in self._canvasLayers]}
        return outData

    def exportImage(self) -> str:
        pw,ph = self._documentSize
        image = ttk.TTkCanvas(width=pw,height=ph)
        for l in self._canvasLayers:
            lx,ly = l.pos()
            l.drawInCanvas(pos=(lx,ly),canvas=image)
        return image.toAnsi()

    def leaveEvent(self, evt):
        self._mouseMove = None
        self._moveData  = None
        self.update()
        return super().leaveEvent(evt)

    @ttk.pyTTkSlot(ToolType)
    def setTool(self, tool):
        self._tool = tool
        self.update()

    def _handleAction(self):
        dx,dy = self._documentPos
        dw,dh = self._documentSize
        ox, oy = self.getViewOffsets()
        mp = self._mousePress
        mm = self._mouseMove
        md = self._mouseDrag
        mr = self._mouseRelease
        l = self._currentLayer
        lx,ly = l.pos()

        if self._tool & ToolType.MOVE and mp and not md:
            if self._tool & ToolType.RESIZE and not md:
                mpx,mpy = mp
                self._resizeData = None
                def _getSelected(_x,_y,_w,_h):
                    _selected = ttk.TTkK.NONE
                    if _x <= mpx < _x+_w and mpy == _y:      _selected |= ttk.TTkK.TOP
                    if _x <= mpx < _x+_w and mpy == _y+_h-1: _selected |= ttk.TTkK.BOTTOM
                    if _y <= mpy < _y+_h and mpx == _x:      _selected |= ttk.TTkK.LEFT
                    if _y <= mpy < _y+_h and mpx == _x+_w-1: _selected |= ttk.TTkK.RIGHT
                    return _selected
                # Main Area Resize Borders
                if selected :=  _getSelected(dx-1,dy-1,dw+2,dh+2):
                    self._resizeData = {'type':PaintArea,'selected':selected,'cb':self.superResize,'geometry':(dx,dy,dw,dh)}
                elif  l:
                    # Selected Layer Resize Borders
                    lx,ly = l.pos()
                    lw,lh = l.size()
                    if selected := _getSelected(dx+lx-1,dy+ly-1,lw+2,lh+2):
                        self._resizeData = {'type':CanvasLayer,'selected':selected,'cb':l.superResize,'geometry':(lx,ly,lw,lh)}
            if not self._resizeData:
                # Get The Layer to Move
                self._moveData = None
                for lm in reversed(self._canvasLayers):
                    mpx,mpy = mp
                    lmx,lmy = lm.pos()
                    self._moveData = {'type':PaintArea,'pos':(dx,dy)}
                    if lm.isOpaque(mpx-lmx-dx,mpy-lmy-dy):
                        self._currentLayer = lm
                        tml = lm
                        self._moveData = {'type':CanvasLayer,'pos':tml.pos(),'layer':tml}
                        self.layerSelected.emit(lm)
                        break

        elif self._tool & ToolType.MOVE and mp and md:
            # Move/Resize Tool
            if self._tool & ToolType.RESIZE and (rData:=self._resizeData):
                _rx,_ry,_rw,_rh = rData['geometry']
                _rdx,_rdy,_rdw,_rdh=(_rx,_ry,_rw,_rh)
                mpx,mpy = mp
                mdx,mdy = md
                diffx = mdx-mpx
                diffy = mdy-mpy
                if rData['selected'] & ttk.TTkK.TOP:    _rdh-=diffy ; _rdy+=diffy
                if rData['selected'] & ttk.TTkK.BOTTOM: _rdh+=diffy
                if rData['selected'] & ttk.TTkK.LEFT:   _rdw-=diffx ; _rdx+=diffx
                if rData['selected'] & ttk.TTkK.RIGHT:  _rdw+=diffx
                rData['cb'](_rdx,_rdy,_rdw,_rdh)
            if not self._resizeData and (mData:=self._moveData):
                mpx,mpy = mp
                mdx,mdy = md
                pdx,pdy = mdx-mpx,mdy-mpy
                if mData['type']==CanvasLayer:
                    px,py = self._moveData['pos']
                    self._moveData['layer'].move(px+pdx,py+pdy)
                    self.layerSelected.emit(self._moveData['layer'])
                elif mData['type']==PaintArea:
                    px,py = self._moveData['pos']
                    self._documentPos = (px+pdx,py+pdy)
            self._retuneGeometry()

        elif self._tool & ToolType.BRUSH:
            if mp and self._tool & ToolType.PICK:
                glbls.brush.setToolType(self._tool & ~ToolType.PICK)
                mpx,mpy = mp
                for lm in reversed(self._canvasLayers):
                    lmx,lmy = lm.pos()
                    if lm.isOpaque(mpx-lmx-dx,mpy-lmy-dy):
                        glbls.brush.setArea(lm.trim().toTTkString())
                        break
                self._mousePress   = None
                self._mouseMove    = None
                self._mouseDrag    = None
                self._mouseRelease = None
            elif self._tool & ToolType.PICK:
                pass # Do not show any preview if we are in picking mode
            elif mp or md:
                if md: mx,my = md
                else:  mx,my = mp
                preview=False
                transparent=self._tool & ToolType.TRANSPARENT
                if self._tool & ToolType.GLYPH:
                    self._currentLayer.placeGlyph(mx-lx-dx,my-ly-dy,self._glyph,self._glyphColor,self._glyphEnabled,preview)
                if self._tool & ToolType.AREA:
                    self._currentLayer.placeArea(mx-lx-dx,my-ly-dy,self._areaBrush,transparent,preview)
            elif mm:
                mx,my = mm
                preview=True
                transparent=self._tool & ToolType.TRANSPARENT
                if self._tool & ToolType.GLYPH:
                    self._currentLayer.placeGlyph(mx-lx-dx,my-ly-dy,self._glyph,self._glyphColor,self._glyphEnabled,preview)
                if self._tool & ToolType.AREA:
                    self._currentLayer.placeArea(mx-lx-dx,my-ly-dy,self._areaBrush,transparent,preview)

        elif self._tool in (ToolType.RECTEMPTY, ToolType.RECTFILL):
            if mr and mp:
                mpx,mpy = mp
                mrx,mry = mr
                preview=False
                self._currentLayer.placeFill((mpx-lx-dx,mpy-ly-dy,mrx-lx-dx,mry-ly-dy),self._tool,self._glyph,self._glyphColor,self._glyphEnabled,preview)
            elif md and mp:
                mpx,mpy = mp
                mrx,mry = md
                preview=True
                self._currentLayer.placeFill((mpx-lx-dx,mpy-ly-dy,mrx-lx-dx,mry-ly-dy),self._tool,self._glyph,self._glyphColor,self._glyphEnabled,preview)
            elif mm:
                mpx,mpy = mm
                mrx,mry = mm
                preview=True
                self._currentLayer.placeFill((mpx-lx-dx,mpy-ly-dy,mrx-lx-dx,mry-ly-dy),self._tool,self._glyph,self._glyphColor,self._glyphEnabled,preview)
        self.update()

    def mouseMoveEvent(self, evt) -> bool:
        ox, oy = self.getViewOffsets()
        self._mouseMove = (evt.x+ox,evt.y+oy)
        self._mouseDrag = None
        self._handleAction()
        return True

    def mouseDragEvent(self, evt) -> bool:
        ox, oy = self.getViewOffsets()
        self._mouseDrag=(evt.x+ox,evt.y+oy)
        self._mouseMove= None
        self._handleAction()
        return True

    def mousePressEvent(self, evt) -> bool:
        ox, oy = self.getViewOffsets()
        self._mousePress=(evt.x+ox,evt.y+oy)
        self._moveData     = None
        self._mouseMove    = None
        self._mouseDrag    = None
        self._mouseRelease = None
        self._handleAction()
        return True

    def mouseReleaseEvent(self, evt) -> bool:
        ox, oy = self.getViewOffsets()
        self._mouseRelease=(evt.x+ox,evt.y+oy)
        self._mouseMove   = None
        self._handleAction()
        self._moveData     = None
        self._resizeData   = None
        self._mousePress   = None
        self._mouseDrag    = None
        self._mouseRelease = None
        return super().mousePressEvent(evt)

    def keyEvent(self, evt) -> bool:
        ret = None
        l = self._currentLayer
        if   l and evt.key == ttk.TTkK.Key_Up:
            x,y = l.pos()
            l.move(x,y-1)
            ret = True
        elif l and evt.key == ttk.TTkK.Key_Down:
            x,y = l.pos()
            l.move(x,y+1)
            ret = True
        elif l and evt.key == ttk.TTkK.Key_Left:
            x,y = l.pos()
            l.move(x-1,y)
            ret = True
        elif l and evt.key == ttk.TTkK.Key_Right:
            x,y = l.pos()
            l.move(x+1,y)
            ret = True
        elif evt.mod==ttk.TTkK.ControlModifier and evt.key == ttk.TTkK.Key_V:
            self.paste()
            ret = True
        elif evt.mod==ttk.TTkK.ControlModifier and evt.key == ttk.TTkK.Key_C:
            if self._currentLayer:
                text = self._currentLayer.toTTkString()
                self.copy(text)
                ret = True
        else:
            return super().keyEvent(evt)
        self._retuneGeometry()
        self.update()
        if ret is None:
            return super().keyEvent(evt)
        return ret

    @ttk.pyTTkSlot()
    def paste(self):
        txt = self._clipboard.text()
        self.pasteEvent(txt)

    @ttk.pyTTkSlot(ttk.TTkString)
    def copy(self, text):
        self._clipboard.setText(text)

    def pasteEvent(self, txt:str):
        l = self.newLayer()
        l.importTTkString(ttk.TTkString(txt))
        self.update()
        return True

    @ttk.pyTTkSlot(ttk.TTkString)
    def setAreaBrush(self, ab:ttk.TTkString):
        self._areaBrush = CanvasLayer()
        self._areaBrush.importTTkString(ab)

    @ttk.pyTTkSlot()
    def updateGlyph(self):
        self.setGlyph(glbls.brush.glyph())
        self.setGlyphColor(glbls.brush.color())
        self._glyphEnabled = glbls.brush.glyphEnabled()

    def glyph(self):
        return self._glyph
    @ttk.pyTTkSlot(str)
    def setGlyph(self, glyph):
        if len(glyph) <= 0: return
        if type(glyph)==str:
            self._glyph = glyph[0]
        if type(glyph)==ttk.TTkString:
            self._glyph = glyph.charAt(0)
            self._glyphColor = glyph.colorAt(0)

    def glyphColor(self):
        return self._glyphColor
    @ttk.pyTTkSlot(ttk.TTkColor)
    def setGlyphColor(self, color):
        self._glyphColor = color

    @ttk.pyTTkSlot(ttk.TTkColor)
    def setTrans(self, color:ttk.TTkColor):
        r,g,b = color.bgToRGB()
        self._transparentColor = {
            'base':color,
            'dim':     ttk.TTkColor.bg(f'#{int(r*0.3):02x}{int(g*0.3):02x}{int(b*0.3):02x}'),
            'layer':   ttk.TTkColor.bg(f'#{int(r*0.6):02x}{int(g*0.6):02x}{int(b*0.6):02x}'),
            'layerDim':ttk.TTkColor.bg(f'#{int(r*0.2):02x}{int(g*0.2):02x}{int(b*0.2):02x}')}
        self.update()

    def paintEvent(self, canvas:ttk.TTkCanvas):
        dx,dy = self._documentPos
        ox, oy = self.getViewOffsets()
        dw,dh = self._documentSize
        dox,doy = dx-ox,dy-oy
        cw,ch = canvas.size()
        w=min(cw,dw)
        h=min(ch,dh)
        tcb = self._transparentColor['base']
        tcd = self._transparentColor['dim']

        if l:=self._currentLayer:
            tclb = self._transparentColor['layer']
            tcld = self._transparentColor['layerDim']
            lx,ly = l.pos()
            lw,lh = l.size()
            canvas.fill(pos=(0     ,ly+doy),size=(cw,lh),color=tcld)
            canvas.fill(pos=(lx+dox,0     ),size=(lw,ch),color=tcld)
            canvas.fill(pos=(lx+dox,ly+doy),size=(lw,lh),color=tclb)
        canvas.fill(pos=(dx-ox,dy-oy),size=(dw,dh),color=tcb)
        canvas.fill(pos=(0    ,dy-oy-1), size=(cw,1),color=tcd)
        canvas.fill(pos=(0    ,dy-oy+dh),size=(cw,1),color=tcd)
        canvas.fill(pos=(dx-ox-2 ,0    ),size=(2,ch),color=tcd)
        canvas.fill(pos=(dx-ox+dw,0    ),size=(2,ch),color=tcd)

        for l in self._canvasLayers:
            lx,ly = l.pos()
            l.drawInCanvas(pos=(lx+dox,ly+doy),canvas=canvas)

        if self._tool & ToolType.RESIZE:
            rd = self._resizeData
            def _drawResizeBorders(_rx,_ry,_rw,_rh,_sel):
                selColor = ttk.TTkColor.YELLOW + ttk.TTkColor.BG_BLUE
                # canvas.drawBox(pos=_pos,size=_size)
                canvas.drawText(pos=(_rx      ,_ry      ),text='─'*_rw, color=selColor if _sel & ttk.TTkK.TOP    else ttk.TTkColor.RST)
                canvas.drawText(pos=(_rx      ,_ry+_rh-1),text='─'*_rw, color=selColor if _sel & ttk.TTkK.BOTTOM else ttk.TTkColor.RST)
                for _y in range(_ry,_ry+_rh):
                    canvas.drawText(pos=(_rx      ,_y),text='│',color=selColor if _sel & ttk.TTkK.LEFT  else ttk.TTkColor.RST)
                    canvas.drawText(pos=(_rx+_rw-1,_y),text='│',color=selColor if _sel & ttk.TTkK.RIGHT else ttk.TTkColor.RST)
                canvas.drawChar(pos=(_rx      ,_ry      ), char='▛')
                canvas.drawChar(pos=(_rx+_rw-1,_ry      ), char='▜')
                canvas.drawChar(pos=(_rx      ,_ry+_rh-1), char='▙')
                canvas.drawChar(pos=(_rx+_rw-1,_ry+_rh-1), char='▟')

            sMain  = rd['selected'] if rd and rd['type'] == PaintArea   else ttk.TTkK.NONE
            sLayer = rd['selected'] if rd and rd['type'] == CanvasLayer else ttk.TTkK.NONE

            _drawResizeBorders(dx-ox-1, dy-oy-1, dw+2, dh+2, sMain)

            if l:=self._currentLayer:
                lx,ly = l.pos()
                lw,lh = l.size()
                _drawResizeBorders(lx+dx-ox-1, ly+dy-oy-1, lw+2, lh+2, sLayer)


class PaintScrollArea(ttk.TTkAbstractScrollArea):
    def __init__(self, pwidget:PaintArea, **kwargs):
        super().__init__(**kwargs)
        self.setViewport(pwidget)


# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['ToolsPanel']

import os

import TermTk as ttk

from .palette      import Palette
from .const        import ToolType
from .glbls        import glbls

class ToolsPanel(ttk.TTkVBoxLayout):
    __slots__ = ('_palette',
                 '_glyph','_color',
                 '_la_brush_g', '_ta_brush_a'
                 # Signals
                 'pickArea', 'toolSelected','areaChanged')
    def __init__(self, *args, **kwargs):
        self.pickArea = ttk.pyTTkSignal()
        self.areaChanged = ttk.pyTTkSignal(ttk.TTkString)
        self.toolSelected = ttk.pyTTkSignal(ToolType)
        self._color = ttk.TTkColor.RST
        super().__init__(*args, **kwargs)
        self._palette  = Palette(maxHeight=12)
        self.addWidget(self._palette)

        # Layout for the toggle buttons
        lToggleFgBg = ttk.TTkHBoxLayout()
        cb_p_fg = ttk.TTkCheckbox(text="-FG-", checked=ttk.TTkK.Checked)
        cb_p_bg = ttk.TTkCheckbox(text="-BG-", checked=ttk.TTkK.Checked)
        cb_p_gl = ttk.TTkCheckbox(text="Glyph", checked=ttk.TTkK.Checked)
        lToggleFgBg.addWidgets([cb_p_fg,cb_p_bg,cb_p_gl])
        lToggleFgBg.addItem(ttk.TTkLayout())
        cb_p_fg.toggled.connect(self._palette.enableFg)
        cb_p_bg.toggled.connect(self._palette.enableBg)
        cb_p_gl.toggled.connect(glbls.brush.setGlyphEnabled)
        self.addItem(lToggleFgBg)

        # Toolset
        lTools = ttk.TTkGridLayout()
        main_toolset = ttk.TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/tools.tui.json"))

        ra_move   = main_toolset.getWidgetByName("ra_move")
        ra_brush  = main_toolset.getWidgetByName("ra_brush")
        ra_line   = main_toolset.getWidgetByName("ra_line")
        ra_rect   = main_toolset.getWidgetByName("ra_rect")
        ra_oval   = main_toolset.getWidgetByName("ra_oval")

        ra_rect_f = main_toolset.getWidgetByName("ra_rect_f")
        ra_rect_e = main_toolset.getWidgetByName("ra_rect_e")

        ra_brush_g = main_toolset.getWidgetByName("ra_brush_g")
        ra_brush_a = main_toolset.getWidgetByName("ra_brush_a")

        cb_move_r = main_toolset.getWidgetByName("cb_move_r")

        la_brush_g = ttk.TTkLabel(   maxHeight=1, minHeight=1)
        btn_pick_a = ttk.TTkButton(  maxHeight=1, minHeight=1, visible=False, text='Pick Area')
        cb_trans_a = ttk.TTkCheckbox(maxHeight=1, minHeight=1, visible=False, text='Transparent')
        ta_brush_a = ttk.TTkTextEdit(                          visible=False, lineNumber=True, readOnly=False)

        spacer = ttk.TTkSpacer()
        # spacer = ttk.TTkFrame(border=True)

        self._la_brush_g = la_brush_g
        self._ta_brush_a = ta_brush_a

        @ttk.pyTTkSlot()
        def _checkTools():
            tool = ToolType.BRUSH | ToolType.GLYPH
            if ra_move.isChecked():
                tool  = ToolType.MOVE
                if cb_move_r.isChecked():
                    tool |= ToolType.RESIZE
            elif ra_brush.isChecked():
                tool = ToolType.BRUSH
                if ra_brush_g.isChecked():
                    tool |= ToolType.GLYPH
                else:
                    tool |= ToolType.AREA
                    if cb_trans_a.isChecked():
                        tool |= ToolType.TRANSPARENT
            elif ra_rect.isChecked():
                if ra_rect_e.isChecked():
                    tool = ToolType.RECTEMPTY
                else:
                    tool = ToolType.RECTFILL

            spacer.show()
            la_brush_g.hide()
            cb_trans_a.hide()
            btn_pick_a.hide()
            ta_brush_a.hide()
            if tool & ToolType.GLYPH:
                la_brush_g.show()
                spacer.show()
            elif tool & ToolType.AREA:
                cb_trans_a.show()
                btn_pick_a.show()
                ta_brush_a.show()
                spacer.hide()

            glbls.brush.setToolType(tool)
            lTools.update()
            self.update()

        @ttk.pyTTkSlot(bool)
        def _emitTool(checked):
            if not checked: return
            _checkTools()

        @ttk.pyTTkSlot()
        def _pick():
            glbls.brush.setToolType(
                ToolType.BRUSH |
                ToolType.AREA |
                ToolType.PICK )

        btn_pick_a.clicked.connect(_pick)

        ra_rect.toggled.connect(ra_rect_f.setEnabled)
        ra_rect.toggled.connect(ra_rect_e.setEnabled)
        ra_move.toggled.connect(cb_move_r.setEnabled)
        ra_brush.toggled.connect(ra_brush_g.setEnabled)
        ra_brush.toggled.connect(ra_brush_a.setEnabled)
        # ra_brush_g.toggled.connect(la_brush_g.setVisible)
        # ra_brush_a.toggled.connect(btn_pick_a.setVisible)
        # ra_brush_a.toggled.connect(ta_brush_a.setVisible)
        ra_brush_g.toggled.connect(self.update)
        ra_brush_a.toggled.connect(self.update)

        ra_move.toggled.connect(   _emitTool)
        ra_brush.toggled.connect(  _emitTool)
        ra_line.toggled.connect(   _emitTool)
        ra_rect.toggled.connect(   _emitTool)
        ra_rect_f.toggled.connect( _emitTool)
        ra_rect_e.toggled.connect( _emitTool)
        ra_oval.toggled.connect(   _emitTool)
        ra_brush_g.toggled.connect(_emitTool)
        ra_brush_a.toggled.connect(_emitTool)
        cb_trans_a.toggled.connect(_checkTools)
        cb_move_r.toggled.connect( _checkTools)

        ta_brush_a.textChanged.connect(lambda : self.areaChanged.emit(ta_brush_a.toRawText()))

        # lTools.addWidget(ra_move    ,0,0)
        # lTools.addWidget(cb_move_r  ,0,1)
        # lTools.addWidget(ra_select  ,1,0)
        # lTools.addWidget(ra_brush   ,2,0)
        # lTools.addWidget(ra_brush_g ,2,1)
        # lTools.addWidget(ra_brush_a ,2,2)
        # lTools.addWidget(ra_line    ,3,0)
        # lTools.addWidget(ra_rect    ,4,0)
        # lTools.addWidget(ra_rect_f  ,4,1)
        # lTools.addWidget(ra_rect_e  ,4,2)
        # lTools.addWidget(ra_oval    ,5,0)
        lTools.addWidget(main_toolset ,1,0,1,3)
        lTools.addWidget(la_brush_g ,6,0)
        lTools.addWidget(btn_pick_a ,7,2,1,1)
        lTools.addWidget(cb_trans_a ,7,0,1,2)
        lTools.addWidget(ta_brush_a ,8,0,1,3)
        lTools.addWidget(spacer     ,9,0,1,3)

        self.addItem(lTools)

        self._palette.colorSelected.connect(glbls.brush.setColor)

        glbls.brush.glyphChanged.connect(self._refreshColor)
        glbls.brush.glyphEnabledChanged.connect(self._refreshColor)
        glbls.brush.areaChanged.connect( self.setArea)
        glbls.brush.colorChanged.connect(self._refreshColor)
        glbls.brush.setColor(self._palette.color())

        _checkTools()
        self._refreshColor()

        # brush
        # line
        # rettangle [empty,fill]
        # oval [empty,fill]

    @ttk.pyTTkSlot()
    def _refreshColor(self):
        color = glbls.brush.color()
        glyph = glbls.brush.glyph()
        self._la_brush_g.setText(
                ttk.TTkString("Glyph: '") +
                ttk.TTkString(glyph,color) +
                ttk.TTkString("'"))

    @ttk.pyTTkSlot(ttk.TTkColor)
    def setColor(self, color:ttk.TTkColor):
        self._refreshColor()

    @ttk.pyTTkSlot(ttk.TTkString)
    def setArea(self, area:ttk.TTkString):
        self._ta_brush_a.setText(area)



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['Layers','LayerData']

import TermTk as ttk

class LayerData():
    __slots__ = ('_name','_data',
                 #signals
                 'nameChanged','visibilityToggled')
    def __init__(self,name:ttk.TTkString=ttk.TTkString('New'),data=None) -> None:
        self._name:ttk.TTkString = ttk.TTkString(name) if type(name)==str else name
        self.visibilityToggled = ttk.pyTTkSignal(bool)
        self._data = data
        self.nameChanged = ttk.pyTTkSignal(str)
    def name(self):
        return self._name
    def setName(self,name):
        self.nameChanged.emit(name)
        self._name = name
    def data(self):
        return self._data
    def setData(self,data):
        self._data = data

class _layerButton(ttk.TTkContainer):
    classStyle = {
                'default':     {'color': ttk.TTkColor.fg("#dddd88")+ttk.TTkColor.bg("#000044"),
                                'borderColor': ttk.TTkColor.fg('#CCDDDD'),
                                'grid':1},
                'disabled':    {'color': ttk.TTkColor.fg('#888888'),
                                'borderColor':ttk.TTkColor.fg('#888888'),
                                'grid':0},
                'hover':       {'color': ttk.TTkColor.fg("#dddd88")+ttk.TTkColor.bg("#000050")+ttk.TTkColor.BOLD,
                                'borderColor': ttk.TTkColor.fg("#FFFFCC")+ttk.TTkColor.BOLD,
                                'grid':1},
                'selected':     {'color': ttk.TTkColor.fg("#dddd88")+ttk.TTkColor.bg("#004488"),
                                'borderColor': ttk.TTkColor.fg("#FFFF00"),
                                'grid':0},
                'unchecked':   {'color': ttk.TTkColor.fg("#dddd88")+ttk.TTkColor.bg("#000044"),
                                'borderColor': ttk.TTkColor.RST,
                                'grid':3},
                'clicked':     {'color': ttk.TTkColor.fg("#FFFFDD")+ttk.TTkColor.BOLD,
                                'borderColor': ttk.TTkColor.fg("#DDDDDD")+ttk.TTkColor.BOLD,
                                'grid':0},
                'focus':       {'color': ttk.TTkColor.fg("#dddd88")+ttk.TTkColor.bg("#000044")+ttk.TTkColor.BOLD,
                                'borderColor': ttk.TTkColor.fg("#ffff00") + ttk.TTkColor.BOLD,
                                'grid':1},
            }

    __slots__ = ('_layerData','_first', '_isSelected', '_layerVisible',
                 '_ledit',
               # signals
               'clicked', 'visibilityToggled',
               )
    def __init__(self, layer:LayerData, **kwargs):
        self.clicked = ttk.pyTTkSignal(_layerButton)
        self._layerData:LayerData = layer
        self._isSelected = False
        self._first = True
        self._layerVisible = True
        self.visibilityToggled = layer.visibilityToggled

        super().__init__(**kwargs|{'layout':ttk.TTkGridLayout()})
        self.setPadding(1,1,7,2)
        self._ledit = ttk.TTkLineEdit(parent=self, text=layer.name(),visible=False)
        self._ledit.focusChanged.connect(self._ledit.setVisible)
        self._ledit.textEdited.connect(self._textEdited)
        # self.setFocusPolicy(ttk.TTkK.ClickFocus)

    @ttk.pyTTkSlot(str)
    def _textEdited(self, text):
        self._layerData.setName(text)

    def mousePressEvent(self, evt) -> bool:
        if evt.x <= 3:
            self._layerVisible = not self._layerVisible
            self.visibilityToggled.emit(self._layerVisible)
        self.setFocus()
        self.update()
        return True

    def mouseReleaseEvent(self, evt) -> bool:
        self.clicked.emit(self)
        return True

    def mouseDoubleClickEvent(self, evt) -> bool:
        self._ledit.setVisible(True)
        self._ledit.setFocus()
        return True

    def mouseDragEvent(self, evt) -> bool:
        drag = ttk.TTkDrag()
        drag.setData(self)
        name = self._layerData.name()
        pm = ttk.TTkCanvas(width=len(name)+4,height=3)
        pm.drawBox(pos=(0,0),size=pm.size())
        pm.drawText(pos=(2,1), text=name)
        drag.setHotSpot(5, 1)
        drag.setPixmap(pm)
        drag.exec()
        return True

    def paintEvent(self, canvas: ttk.TTkCanvas):
        if self._isSelected:
            style = self.style()['selected']
        else:
            style = self.currentStyle()
        borderColor = style['borderColor']
        textColor   = style['color']
        btnVisible = '▣' if self._layerVisible else '□'
        w,h = self.size()
        canvas.drawText(    pos=(0,0),text=f"     ┏{'━'*(w-7)}┓",color=borderColor)
        canvas.drawText(    pos=(0,2),text=f"     ┗{'━'*(w-7)}┛",color=borderColor)
        if self._first:
            canvas.drawText(pos=(0,1),text=f" {btnVisible} - ┃{' '*(w-7)}┃",color=borderColor)
        else:
            canvas.drawText(pos=(0,1),text=f" {btnVisible} - ╽{' '*(w-7)}╽",color=borderColor)
        canvas.drawTTkString(pos=(7,1),text=self._layerData.name(), width=w-9, color=textColor)

class LayerScrollWidget(ttk.TTkAbstractScrollView):
    __slots__ = ('_layers','_selected', '_dropTo',
                 # Signals
                 'layerSelected','layerAdded','layerDeleted','layersOrderChanged')
    def __init__(self, **kwargs):
        self.layerSelected = ttk.pyTTkSignal(LayerData)
        self.layerAdded    = ttk.pyTTkSignal(LayerData)
        self.layerDeleted  = ttk.pyTTkSignal(LayerData)
        self.layersOrderChanged = ttk.pyTTkSignal(list[LayerData])

        self._selected = None
        self._dropTo = None
        self._layers:list[_layerButton] = []
        super().__init__(**kwargs)
        self.viewChanged.connect(self._placeTheButtons)
        self.viewChanged.connect(self._viewChangedHandler)

    @ttk.pyTTkSlot()
    def _viewChangedHandler(self):
        x,y = self.getViewOffsets()
        self.layout().setOffset(-x,-y)

    def viewFullAreaSize(self) -> tuple:
        _,_,w,h = self.layout().fullWidgetAreaGeometry()
        return w,h

    def viewDisplayedSize(self) -> tuple:
        return self.size()

    def maximumWidth(self):   return 0x10000
    def maximumHeight(self):  return 0x10000
    def minimumWidth(self):   return 0
    def minimumHeight(self):  return 0

    def selectLayerByData(self, data):
        if sel:=self._selected:
            sel._isSelected = False
            sel.update()
        for layBtn in self._layers:
            if layBtn._layerData.data() == data:
                self._selected = layBtn
                layBtn._isSelected = True
        self.update()

    @ttk.pyTTkSlot(_layerButton)
    def _clickedLayer(self, layerButton:_layerButton):
        if sel:=self._selected:
            sel._isSelected = False
            sel.update()
        self._selected = layerButton
        layerButton._isSelected = True
        self.layerSelected.emit(layerButton._layerData)
        self.update()

    def clear(self):
        for layBtn in self._layers:
            self.layout().removeWidget(layBtn)
            layBtn.clicked.clear()
            layBtn.visibilityToggled.clear()
            layBtn._layerData.nameChanged.clear()
        self._layers.clear()
        self.update()

    @ttk.pyTTkSlot()
    def moveUp(self):
        return self._moveButton(-1)

    @ttk.pyTTkSlot()
    def moveDown(self):
        return self._moveButton(+1)

    def _moveButton(self,direction):
        if not self._selected: return
        index = self._layers.index(self._selected)
        if index+direction < 0: return
        l = self._layers.pop(index)
        self._layers.insert(index+direction,l)
        self._placeTheButtons()
        self.layersOrderChanged.emit([_l._layerData for _l in self._layers])

    @ttk.pyTTkSlot()
    def addLayer(self,name=None, data=None):
        name = name if name else f"Layer #{len(self._layers)}"
        _l=LayerData(name=name,data=data)
        newLayerBtn:_layerButton  = _layerButton(parent=self,layer=_l)
        self._layers.insert(0,newLayerBtn)
        if sel:=self._selected: sel._isSelected = False
        self._selected = newLayerBtn
        newLayerBtn._isSelected = True
        newLayerBtn.clicked.connect(self._clickedLayer)
        self.viewChanged.emit()
        self._placeTheButtons()
        self.layerAdded.emit(newLayerBtn._layerData)
        return _l

    def _placeTheButtons(self):
        w,h = self.size()
        for i,l in enumerate(self._layers):
            l._first = i==0
            l.setGeometry(0,i*2,w,3)
            l.lowerWidget()
        self.update()

    @ttk.pyTTkSlot()
    def delLayer(self):
        self._layers.remove()

    def dragEnterEvent(self, evt) -> bool:
        if type(evt.data())!=_layerButton: return False
        x,y = self.getViewOffsets()
        self._dropTo = max(0,min(len(self._layers),(evt.y-1+y)//2))
        self.update()
        return True
    def dragLeaveEvent(self, evt) -> bool:
        if type(evt.data())!=_layerButton: return False
        self._dropTo = None
        self.update()
        return True
    def dragMoveEvent(self, evt) -> bool:
        if type(evt.data())!=_layerButton: return False
        x,y = self.getViewOffsets()
        self._dropTo = max(0,min(len(self._layers),(evt.y-1+y)//2))
        self.update()
        ttk.TTkLog.debug(f"{evt.x},{evt.y-y} - {len(self._layers)} - {self._dropTo}")
        return True
    def dropEvent(self, evt) -> bool:
        if type(evt.data())!=_layerButton: return False
        x,y = self.getViewOffsets()
        self._dropTo = None
        data = evt.data()
        # dropPos = len(self._layers)-(evt.y-1)//2
        dropPos = max(0,min(len(self._layers),(evt.y-1+y)//2))
        ttk.TTkLog.debug(f"{evt.x},{evt.y-y} - {len(self._layers)} - {self._dropTo} {dropPos}")
        if dropPos > self._layers.index(data):
            dropPos -= 1
        self._layers.remove(data)
        self._layers.insert(dropPos,data)
        self._placeTheButtons()
        self.layersOrderChanged.emit([_l._layerData for _l in self._layers])
        return True

    # Stupid hack to paint on top of the child widgets
    def paintChildCanvas(self):
        super().paintChildCanvas()
        offx, offy = self.getViewOffsets()
        if self._dropTo == None: return
        canvas = self.getCanvas()
        w,h = canvas.size()
        color = ttk.TTkColor.YELLOW
        canvas.drawText(pos=(0,(self._dropTo)*2-offy),text=f"╞{'═'*(w-2)}╍",color=color)



class Layers(ttk.TTkGridLayout):
    __slots__ = ('_scrollWidget',
                 # Forward Methods
                 'addLayer','clear','selectLayerByData',
                 # Forward Signals
                 'layerSelected','layerAdded','layerDeleted','layersOrderChanged')
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._scrollWidget = _lsw = LayerScrollWidget()
        _sa = ttk.TTkAbstractScrollArea(scrollWidget=self._scrollWidget,minWidth=16)
        _sa.setViewport(_lsw)
        self.addWidget(_sa,0,0,1,5)
        self.addWidget(btnAdd :=ttk.TTkButton(text='add')           ,1,0)
        self.addWidget(btnUp  :=ttk.TTkButton(text='▲',maxWidth=3)  ,1,1)
        self.addWidget(btnDown:=ttk.TTkButton(text='▼',maxWidth=3)  ,1,2)
        # self.addItem(ttk.TTkLayout(),1,3)
        self.addWidget(btnDel :=ttk.TTkButton(text=ttk.TTkString('del',ttk.TTkColor.RED),maxWidth=5),1,4)

        btnAdd.setToolTip( "Create a new Layer\nand add it to the image")
        btnDel.setToolTip( "Delete the selected Layer")
        btnUp.setToolTip(  "Raise the selected Layer one step")
        btnDown.setToolTip("Lower the selected Layer one step")

        btnAdd.clicked.connect( _lsw.addLayer)
        btnDel.clicked.connect( _lsw.delLayer)
        btnUp.clicked.connect(  _lsw.moveUp)
        btnDown.clicked.connect(_lsw.moveDown)

        # forward signals
        self.layerSelected = _lsw.layerSelected
        self.layerSelected = _lsw.layerSelected
        self.layerAdded    = _lsw.layerAdded
        self.layerDeleted  = _lsw.layerDeleted
        self.layersOrderChanged = _lsw.layersOrderChanged

        # forward methods
        self.addLayer = _lsw.addLayer
        self.clear    = _lsw.clear
        self.selectLayerByData = _lsw.selectLayerByData



from .main         import *
from .maintemplate import *
from .toolspanel   import *
from .paintarea    import *
from .painttoolkit import *
from .textarea     import *
from .palette      import *
from .layers       import *
from .canvaslayer  import *
from .about        import *
from .brush        import *
from .glbls        import *
from .const        import *
from .filters      import *
from .importimage  import *
from .exportarea   import *



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['PaintToolKit']

import os

import TermTk as ttk

from .paintarea import *
from .glbls     import glbls

class PaintToolKit(ttk.TTkContainer):
    __slots__ = ('_rSelect', '_rPaint', '_lgliph',
                 '_cbFg', '_cbBg',
                 '_bpFg', '_bpBg', '_bpDef',
                 '_sbDx','_sbDy','_sbDw','_sbDh',
                 '_sbLx','_sbLy','_sbLw','_sbLh',
                 #Signals
                 'updatedTrans')
    def __init__(self, *args, **kwargs):
        ttk.TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/paintToolKit.tui.json"),self)
        self.updatedTrans = ttk.pyTTkSignal(ttk.TTkColor)
        self._lgliph  = self.getWidgetByName("lglyph")
        self._cbFg    = self.getWidgetByName("cbFg")
        self._cbBg    = self.getWidgetByName("cbBg")
        self._bpFg    = self.getWidgetByName("bpFg")
        self._bpBg    = self.getWidgetByName("bpBg")
        self._bpDef   = self.getWidgetByName("bpDef")

        self._sbDx = self.getWidgetByName("sbDx")
        self._sbDy = self.getWidgetByName("sbDy")
        self._sbDw = self.getWidgetByName("sbDw")
        self._sbDh = self.getWidgetByName("sbDh")
        self._sbLx = self.getWidgetByName("sbLx")
        self._sbLy = self.getWidgetByName("sbLy")
        self._sbLw = self.getWidgetByName("sbLw")
        self._sbLh = self.getWidgetByName("sbLh")

        self._bpDef.setColor(ttk.TTkColor.bg('#FF00FF'))
        self._cbFg.toggled.connect(self._refreshColor)
        self._cbBg.toggled.connect(self._refreshColor)

        self._bpFg.colorSelected.connect(self._refreshColor)
        self._bpBg.colorSelected.connect(self._refreshColor)
        self._bpDef.colorSelected.connect(self.updatedTrans.emit)

        glbls.brush.glyphChanged.connect(   lambda:self._refreshColor(False))
        glbls.brush.colorChanged.connect(   self.setColor)

        self._refreshColor(emit=False)

    @ttk.pyTTkSlot(CanvasLayer)
    def updateLayer(self, layer:CanvasLayer):
        lx,ly = layer.pos()
        lw,lh = layer.size()
        self._sbLx.setValue(lx)
        self._sbLy.setValue(ly)
        self._sbLw.setValue(lw)
        self._sbLh.setValue(lh)

    @ttk.pyTTkSlot()
    def _refreshColor(self, emit=True):
        glyph = glbls.brush.glyph()
        color = self.color()
        self._lgliph.setText(
                ttk.TTkString("Glyph: '") +
                ttk.TTkString(glyph,color) +
                ttk.TTkString("'"))
        if emit:
            glbls.brush.setColor(color)

    def color(self):
        color = ttk.TTkColor()
        if self._cbFg.checkState() == ttk.TTkK.Checked:
            color += self._bpFg.color().invertFgBg()
        if self._cbBg.checkState() == ttk.TTkK.Checked:
           color += self._bpBg.color()
        return color

    @ttk.pyTTkSlot(ttk.TTkColor)
    def setColor(self, color:ttk.TTkColor):
        if fg := color.foreground():
            self._cbFg.setCheckState(ttk.TTkK.Checked)
            self._bpFg.setEnabled()
            self._bpFg.setColor(fg.invertFgBg())
        else:
            self._cbFg.setCheckState(ttk.TTkK.Unchecked)
            self._bpFg.setDisabled()

        if bg := color.background():
            self._cbBg.setCheckState(ttk.TTkK.Checked)
            self._bpBg.setEnabled()
            self._bpBg.setColor(bg)
        else:
            self._cbBg.setCheckState(ttk.TTkK.Unchecked)
            self._bpBg.setDisabled()
        self._refreshColor(emit=False)



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['ExportArea']

import TermTk as ttk

from .paintarea    import PaintArea

class ExportArea(ttk.TTkGridLayout):
    __slots__ = ('_paintArea', '_te','_cbCrop', '_cbFull', '_cbPal')
    def __init__(self, paintArea:PaintArea, **kwargs):
        self._paintArea:PaintArea = paintArea
        super().__init__(**kwargs)
        self._te = ttk.TTkTextEdit(lineNumber=True, readOnly=False)
        btn_exIm   = ttk.TTkButton(text="Export Image")
        btn_exLa   = ttk.TTkButton(text="Export Layer")
        btn_exPr   = ttk.TTkButton(text="Export Document")
        self._cbCrop = ttk.TTkCheckbox(text="Crop",checked=True)
        self._cbFull = ttk.TTkCheckbox(text="Full",checked=True)
        self._cbPal  = ttk.TTkCheckbox(text="Palette",checked=True)
        self.addWidget(btn_exLa  ,0,0)
        self.addWidget(btn_exIm  ,0,1)
        self.addWidget(btn_exPr  ,0,2)
        self.addWidget(self._cbCrop,0,3)
        self.addWidget(self._cbFull,0,4)
        self.addWidget(self._cbPal ,0,5)
        self.addWidget(self._te,1,0,1,7)

        btn_exLa.clicked.connect(self._exportLayer)
        btn_exPr.clicked.connect(self._exportDocument)
        btn_exIm.clicked.connect(self._exportImage)

    @ttk.pyTTkSlot()
    def _exportImage(self):
        import base64, gzip
        crop    = self._cbCrop.isChecked()
        palette = self._cbPal.isChecked()
        full    = self._cbFull.isChecked()
        image = self._paintArea.exportImage()
        # trim the lines
        imagexxx = [line.rstrip() for line in image.split("\n")]
        image = "\n".join(imagexxx)
        self._te.setText(image)

        self._te.append("\n")
        self._te.append(ttk.TTkString("# Python Code",ttk.TTkColor.GREEN))
        lines = [f'     "{line}\\n"' for line in image.replace("\033","\\033").split('\n')]
        mergedLines = "\n".join(lines)
        self._te.append(f"image = (\n{mergedLines})")
        self._te.append('print(image)')

        self._te.append("\n")
        self._te.append(ttk.TTkString("# Python Code using pyTermTk archive",ttk.TTkColor.GREEN))
        self._te.append('from TermTk import TTkUtil\n')
        self._te.append('image = TTkUtil.base64_deflate_2_obj(')
        b64str = ttk.TTkUtil.obj_inflate_2_base64(image)
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '")'
        self._te.append(b64list)
        self._te.append('print(image)')

        self._te.append("\n")
        self._te.append(ttk.TTkString("# Python Code using standard archive libs",ttk.TTkColor.GREEN))
        b64str = base64.b64encode(gzip.compress(bytearray(image,encoding='utf8'))).decode("ascii")
        self._te.append('import zlib, pickle, base64\n')
        self._te.append('image = pickle.loads(zlib.decompress(base64.b64decode((')
        b64str = ttk.TTkUtil.obj_inflate_2_base64(image)
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '").encode("ascii"))))'
        self._te.append(b64list)
        self._te.append('print(image)')

        self._te.append("\n")
        self._te.append(ttk.TTkString("# Bash Code",ttk.TTkColor.GREEN))
        lines = [f'echo -e "{line}"' for line in image.replace("\033","\\033").split('\n')]
        mergedLines = "\n".join(lines)
        self._te.append(mergedLines)

        self._te.append("\n")
        self._te.append(ttk.TTkString("# Bash Code Compressed",ttk.TTkColor.GREEN))
        b64str = base64.b64encode(gzip.compress(bytearray(image,encoding='utf8'))).decode("ascii")
        b64list = "\n".join([b64str[i:i+128] for i in range(0,len(b64str),128)])
        self._te.append(f'echo "{b64list}" | base64 -d | zcat\n')

    @ttk.pyTTkSlot()
    def _exportLayer(self):
        crop    = self._cbCrop.isChecked()
        palette = self._cbPal.isChecked()
        full    = self._cbFull.isChecked()
        dd = self._paintArea.exportLayer(full=full,palette=palette,crop=crop)
        if not dd:
            self._te.setText('# No Data toi be saved!!!')
            return

        self._te.setText('# Compressed Data:')
        self._te.append('data = TTkUtil.base64_deflate_2_obj(')
        b64str = ttk.TTkUtil.obj_inflate_2_base64(dd)
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '")'
        self._te.append(b64list)

        self._te.append('\n# Uncompressed Data:')
        outTxt = '{\n'
        for i in dd:
            if i in ('data','colors','palette'): continue
            if type(dd[i]) == str:
                outTxt += f"  '{i}':'{dd[i]}',\n"
            else:
                outTxt += f"  '{i}':{dd[i]},\n"
        outTxt += "    'data':[\n"
        for l in dd['data']:
            outTxt += f"    {l},\n"
        outTxt += "  ],'colors':[\n"
        for l in dd['colors']:
            outTxt += f"    {l},\n"
        if 'palette' in dd:
            outTxt += "  ],'palette':["
            for i,l in enumerate(dd['palette']):
                if not i%10:
                    outTxt += f"\n    "
                outTxt += f"{l},"
        outTxt += "]}\n"
        self._te.append(outTxt)

    @ttk.pyTTkSlot()
    def _exportDocument(self):
        crop    = self._cbCrop.isChecked()
        palette = self._cbPal.isChecked()
        full    = self._cbFull.isChecked()
        dd = self._paintArea.exportDocument(full=full,palette=palette,crop=crop)
        if not dd:
            self._te.setText('# No Data to be saved!!!')
            return

        self._te.setText('# Compressed Data:')
        self._te.append('data = TTkUtil.base64_deflate_2_obj(')
        b64str = ttk.TTkUtil.obj_inflate_2_base64(dd)
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '")'
        self._te.append(b64list)

        self._te.append('\n# Uncompressed Data:')
        outTxt = '{\n'
        for i in dd:
            if i=='layers': continue
            if type(dd[i]) == str:
                outTxt += f"  '{i}':'{dd[i]}',\n"
            else:
                outTxt += f"  '{i}':{dd[i]},\n"
        outTxt +=  "  'layers':[\n"
        for l in dd['layers']:
            outTxt += f"    {l},\n"
        outTxt += "]}\n"
        self._te.append(outTxt)


# MIT License
#
# Copyright (c) 2021 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['TextArea']

import TermTk as ttk

from .glbls import glbls

_StartingText = '''Ansi Editor:

─ ┌─┬┐  ┏━┳┓  ┎─┰┒  ┍━┯┑  ┏┭┲━┱┮┓
━ │ ││  ┃ ┃┃  ┃ ┃┃  │ ││  ┞╁╀┰╀╁┦
│ ├─┼┤  ┣━╋┫  ┠─╂┨  ┝━┿┥  ┟╄╈╇╈╃┧
┃ └─┴┘  ┗━┻┛  ┖─┸┚  ┕━┷┙  ┡┽╊┷╉┾┩
┄    ╴ ╵╷ ╸ ╹╻ ╼ ╽╿       ╽╵┞╼┩┕┙╻
┅    ╶    ╺    ╾          ╿╷┝╾┤┌┮╋╸
┆ ┌─┬┐  ╔═╦╗  ╓─╥╖  ╒═╤╕  ┢┽╆━╅┾┪╹
┇ │ ││  ║ ║║  ║ ║║  │ ││  ┗┵┺━┹┶┛
┈ ├─┼┤  ╠═╬╣  ╟─╫╢  ╞═╪╡  ┞ ┦ ┭ ┵ ┽ ╃ ╇
┉ └─┴┘  ╚═╩╝  ╙─╨╜  ╘═╧╛  ┟ ┧ ┮ ┶ ┾ ╄ ╈
┊ ╭───────────────────╮   ┡ ┩ ┱ ┹ ╀ ╅ ╉
┋ │  ╔═══╗ Some Text  │░  ┢ ┪ ┲ ┺ ╁ ╆ ╊
╌ │  ╚═╦═╝ in the box │▒
╍ ╞═╤══╩══╤═══════════╡▒      ╭─────╮
╎ │ ├──┬──┤           │▒    ╭ │╲ ╳ ╱│
╏ │ ╰──┴──╯           │▓    ╮ │ ╳╳╳ │
═ └───────────────────┘▓    ╯ ╰─────╯
║  ░░░▒▒▒▒▒▒▒▒▒▒▒▒▒▓▓▓▓█    ╰  ╱ ╲ ╳

▄ ▀ █ <- Half,Full
▁ ▂ ▃ ▄ ▅ ▆ ▇ █     Spaces: "    "

█ Full █ ▌ ▐ ▏ ▕  ▔ ▁
▉        ░ ▒ ▓ █
▊
▋        ▘ ▚ ▟
▌        ▝ ▞ ▙
▍        ▖ ▌ ▜
▎        ▗ ▐ ▛
▏ 1/8
  ◢ ◣ ◤ ◥  ■ □ ▢ ▣

Symbols for Legacy Computing
                     🬼   🭀
🬼 🭇 🭗 🭢  🭨  🭍 🭑 🬽  🭌 🬿 🭐 🭎
🬽 🭈 🭘 🭣  🭩
🬾 🭉 🭙 🭤  🭪
🬿 🭊 🭚 🭥  🭫
🭀 🭋 🭛 🭦  🭬
🭌 🭁 🭝 🭒  🭭
🭍 🭂 🭞 🭓  🭮
🭎 🭃 🭟 🭔  🭯
🭏 🭄 🭠 🭕  🮚
🭐 🭅 🭡 🭖  🮛
🭑 🭆 🭜 🭧

    🭇🬼
    🭃🭌🬿
    🭥🭒█🭏🬼
 🭋🭍🭑🬽🭢🭕█🭌🬿
 🭅███🭀 🭥🭒█🭏🬼
🭋████🭐  🭢🭕█🭌🬿
🭅██🭞🭜🭘 🭈🭆🭂███🭏🬼
🭣🭧🭚 🭈🭆🭂██████🭠🭗
       🭥🭒██🭝🭚
        🭢🭕🭠🭗

'''
class TextArea(ttk.TTkGridLayout):
    __slots__ = ('_te')
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._te = ttk.TTkTextEdit(lineNumber=True, readOnly=False)
        self._te.setText(_StartingText)

        self.addItem(wrapLayout := ttk.TTkGridLayout(), 0,0)
        self.addItem(fontLayout := ttk.TTkGridLayout(columnMinWidth=1), 1,0)
        self.addWidget(self._te,2,0)

        wrapLayout.addWidget(ttk.TTkLabel(text="Wrap: ", maxWidth=6),0,0)
        wrapLayout.addWidget(lineWrap := ttk.TTkComboBox(list=['NoWrap','WidgetWidth','FixedWidth'], maxWidth=20),0,1)
        wrapLayout.addWidget(ttk.TTkLabel(text=" Type: ",maxWidth=7),0,2)
        wrapLayout.addWidget(wordWrap := ttk.TTkComboBox(list=['WordWrap','WrapAnywhere'], maxWidth=20, enabled=False),0,3)
        wrapLayout.addWidget(ttk.TTkLabel(text=" FixW: ",maxWidth=7),0,4)
        wrapLayout.addWidget(fixWidth := ttk.TTkSpinBox(value=self._te.wrapWidth(), maxWidth=5, maximum=500, minimum=10, enabled=False),0,5)
        wrapLayout.addWidget(ttk.TTkSpacer(),0,10)

        # Empty columns/cells are 1 char wide due to "columnMinWidth=1" parameter in the GridLayout
        #           1       3                    8                11
        #    0       2       4    5    6    7     9       10       12
        # 0  [ ] FG  [ ] BG  [ ] LineNumber
        # 1  ┌─────┐ ┌─────┐ ┌──────┐┌──────┐
        # 2  │     │ │     │ │ UNDO ││ REDO │
        # 3  └─────┘ └─────┘ ╘══════╛└──────┘ ┕━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┙

        # Char Fg/Bg buttons
        fontLayout.addWidget(cb_fg := ttk.TTkCheckbox(text=" FG"),0,0)
        fontLayout.addWidget(btn_fgColor := ttk.TTkColorButtonPicker(border=True, enabled=False, maxSize=(7,3)),1,0)

        fontLayout.addWidget(cb_bg := ttk.TTkCheckbox(text=" BG"),0,2)
        fontLayout.addWidget(btn_bgColor := ttk.TTkColorButtonPicker(border=True, enabled=False, maxSize=(7   ,3)),1,2)

        fontLayout.addWidget(cb_linenumber := ttk.TTkCheckbox(text=" LineNumber", checked=True),0,4,1,3)

        # Undo/Redo buttons
        fontLayout.addWidget(btn_undo := ttk.TTkButton(border=True, maxSize=(8,3), enabled=self._te.isUndoAvailable(), text=' UNDO '),1,4)
        fontLayout.addWidget(btn_redo := ttk.TTkButton(border=True, maxSize=(8,3), enabled=self._te.isRedoAvailable(), text=' REDO '),1,5)
        # Undo/Redo events
        self._te.undoAvailable.connect(btn_undo.setEnabled)
        self._te.redoAvailable.connect(btn_redo.setEnabled)
        btn_undo.clicked.connect(self._te.undo)
        btn_redo.clicked.connect(self._te.redo)

        @ttk.pyTTkSlot(ttk.TTkColor)
        def _currentColorChangedCB(format):
            if fg := format.foreground():
                cb_fg.setCheckState(ttk.TTkK.Checked)
                btn_fgColor.setEnabled()
                btn_fgColor.setColor(fg.invertFgBg())
            else:
                cb_fg.setCheckState(ttk.TTkK.Unchecked)
                btn_fgColor.setDisabled()

            if bg := format.background():
                cb_bg.setCheckState(ttk.TTkK.Checked)
                btn_bgColor.setEnabled()
                btn_bgColor.setColor(bg)
            else:
                cb_bg.setCheckState(ttk.TTkK.Unchecked)
                btn_bgColor.setDisabled()

        self._te.currentColorChanged.connect(_currentColorChangedCB)

        def _setStyle():
            color = ttk.TTkColor()
            if cb_fg.checkState() == ttk.TTkK.Checked:
                color += btn_fgColor.color().invertFgBg()
            if cb_bg.checkState() == ttk.TTkK.Checked:
                color += btn_bgColor.color()
            cursor = self._te.textCursor()
            cursor.applyColor(color)
            cursor.setColor(color)
            self._te.setFocus()

        cb_fg.toggled.connect(btn_fgColor.setEnabled)
        cb_bg.toggled.connect(btn_bgColor.setEnabled)
        cb_fg.clicked.connect(_setStyle)
        cb_bg.clicked.connect(_setStyle)

        cb_linenumber.toggled.connect(self._te.setLineNumber)

        btn_fgColor.colorSelected.connect(_setStyle)
        btn_bgColor.colorSelected.connect(_setStyle)

        lineWrap.setCurrentIndex(0)
        wordWrap.setCurrentIndex(1)

        fixWidth.valueChanged.connect(self._te.setWrapWidth)

        @ttk.pyTTkSlot(int)
        def _lineWrapCallback(index):
            if index == 0:
                self._te.setLineWrapMode(ttk.TTkK.NoWrap)
                wordWrap.setDisabled()
                fixWidth.setDisabled()
            elif index == 1:
                self._te.setLineWrapMode(ttk.TTkK.WidgetWidth)
                wordWrap.setEnabled()
                fixWidth.setDisabled()
            else:
                self._te.setLineWrapMode(ttk.TTkK.FixedWidth)
                self._te.setWrapWidth(fixWidth.value())
                wordWrap.setEnabled()
                fixWidth.setEnabled()

        lineWrap.currentIndexChanged.connect(_lineWrapCallback)

        @ttk.pyTTkSlot(int)
        def _wordWrapCallback(index):
            if index == 0:
                self._te.setWordWrapMode(ttk.TTkK.WordWrap)
            else:
                self._te.setWordWrapMode(ttk.TTkK.WrapAnywhere)

        wordWrap.currentIndexChanged.connect(_wordWrapCallback)

        self._te.document().cursorPositionChanged.connect(self._cursorPositionChanged)

    @ttk.pyTTkSlot(ttk.TTkTextCursor)
    def _cursorPositionChanged(self, cursor):
        ch = cursor.positionChar()
        # color = cursor.positionColor()
        glbls.brush.setGlyph(ch)

    def te(self):
        return self._te



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['main']

import argparse

import TermTk as ttk

from .maintemplate import PaintTemplate

ttk.TTkTheme.loadTheme(ttk.TTkTheme.NERD)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', type=str, nargs='?', help='the file to open')
    # parser.add_argument('-k', '--showkeys', action='store_true', help='display the keypresses and mouse interactions')
    args = parser.parse_args()

    root = ttk.TTk(
            title="Dumb Paint Tool",
            layout=ttk.TTkGridLayout(),
            mouseTrack=True,
            sigmask=(
                ttk.TTkTerm.Sigmask.CTRL_C |
                ttk.TTkTerm.Sigmask.CTRL_Q |
                ttk.TTkTerm.Sigmask.CTRL_S |
                ttk.TTkTerm.Sigmask.CTRL_Z ))

    PaintTemplate(parent=root,fileName=args.filename)

    root.mainloop()

if __name__ == '__main__':
    main()


# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['CanvasLayer']

import TermTk as ttk

from .const import ToolType

# Canvas Layer structure
# The data may include more areas than the visible one
# This is helpful in case of resize to not lose the drawn areas
#
#   |---| OffsetX
#       x
#   ╭────────────────╮    -             -
#   │                │    | OffsetY     |
# y │   ┌───────┐    │ \  -             | Data
#   │   │Visible│    │ | h              |
#   │   └───────┘    │ /                |
#   │                │                  |
#   └────────────────┘                  -
#        \---w--/

class CanvasLayer():
    __slot__ = ('_pos','_name','_visible','_size','_data','_colors','_preview','_offset',
                'changed')
    def __init__(self) -> None:
        self.changed = ttk.pyTTkSignal()
        self._pos  = (0,0)
        self._size = (0,0)
        self._offset = (0,0)
        self._name = ""
        self._visible = True
        self._preview = None
        self._data:  list[list[str         ]] = []
        self._colors:list[list[ttk.TTkColor]] = []

    def update(self):
        self.changed.emit()

    def pos(self):
        return self._pos
    def size(self):
        return self._size

    def visible(self):
        return self._visible
    @ttk.pyTTkSlot(bool)
    def setVisible(self, visible):
        self._visible = visible

    def name(self):
        return self._name
    @ttk.pyTTkSlot(str)
    def setName(self, name):
        self._name = name

    def isOpaque(self,x,y):
        if not self._visible: return False
        ox,oy = self._offset
        w,h = self._size
        data = self._data
        colors = self._colors
        if 0<=x<w and 0<=y<h:
            return data[oy+y][ox+x] != ' ' or colors[oy+y][ox+x].background()
        return False

    def move(self,x,y):
        self._pos=(x,y)
        self.changed.emit()

    def resize(self,w,h):
        self._size = (w,h)
        self._data   = (self._data   + [[] for _ in range(h)])[:h]
        self._colors = (self._colors + [[] for _ in range(h)])[:h]
        for i in range(h):
            self._data[i]   = (self._data[i]   + [' '              for _ in range(w)])[:w]
            self._colors[i] = (self._colors[i] + [ttk.TTkColor.RST for _ in range(w)])[:w]
        self.changed.emit()

    def superResize(self,dx,dy,dw,dh):
        ox,oy = self._offset
        x,y = self.pos()
        w,h = self.size()
        daw = len(self._data[0])
        dah = len(self._data)
        diffx = dx-x
        diffy = dy-y
        self._preview = None
        if ox < x-dx: # we need to resize and move ox
            _nw = x-dx-ox
            ox  = x-dx
            self._data   = [([' ']*_nw             ) + _r for _r in self._data]
            self._colors = [([ttk.TTkColor.RST]*_nw) + _r for _r in self._colors]
        if dw+ox > daw:
            _nw = dw+ox-daw
            self._data   = [_r + ([' ']*_nw             )  for _r in self._data]
            self._colors = [_r + ([ttk.TTkColor.RST]*_nw)  for _r in self._colors]
        daw = len(self._data[0])
        if oy < y-dy: # we need to resize and move ox
            _nh = y-dy-oy
            oy  = y-dy
            self._data   = [[' ']*daw              for _ in range(_nh)] + self._data
            self._colors = [[ttk.TTkColor.RST]*daw for _ in range(_nh)] + self._colors
        if dh+oy > dah:
            _nh = dh+oy-dah
            self._data   = self._data   + [[' ']*daw              for _ in range(_nh)]
            self._colors = self._colors + [[ttk.TTkColor.RST]*daw for _ in range(_nh)]
        self._offset = (ox+diffx,oy+diffy)
        self._pos  = (dx,dy)
        self._size = (dw,dh)
        self.changed.emit()

    def clean(self):
        w,h = self._size
        self._offset = (0,0)
        self._preview = None
        for i in range(h):
            self._data[i]   = [' ']*w
            self._colors[i] = [ttk.TTkColor.RST]*w
        self.changed.emit()

    def toTTkString(self):
        ret = []
        pw,ph = self._size
        ox,oy = self._offset
        if not (pw and ph) : return ttk.TTkString()
        for d,c in zip(self._data[oy:oy+ph],self._colors[oy:oy+ph]):
            ret.append(ttk.TTkString._importString1(''.join(d[ox:ox+pw]),c[ox:ox+pw]))
        return ttk.TTkString('\n').join(ret)

    def exportLayer(self, full=False, palette=True, crop=True):
        #           xa|----------|  xb
        #     px        |-----------| = max(px,px+xa-ox)
        # Offset |------|    pw
        #   Data |----------------------------|
        #             daw

        # Don't try this at home
        ox,oy = self._offset
        px,py  = self.pos()
        pw,ph  = self.size()

        if full:
            data   = self._data
            colors = self._colors
        else:
            data   = [row[ox:ox+pw] for row in self._data[  oy:oy+ph]  ]
            colors = [row[ox:ox+pw] for row in self._colors[oy:oy+ph]  ]
            ox=oy=0

        daw = len(data[0])
        dah = len(data)

        # get the bounding box
        if crop:
            xa,xb,ya,yb = 0x10000,0,0x10000,0
            for y,(drow,crow) in enumerate(zip(data,colors)):
                for x,(d,c) in enumerate(zip(drow,crow)):
                    if d != ' ' or c.background():
                        xa,xb = min(x,xa),max(x,xb)
                        ya,yb = min(y,ya),max(y,yb)
            if (xa,xb,ya,yb) == (0x10000,0,0x10000,0):
                xa=xb=ya=yb=0
        else:
            xa,xb,ya,yb = 0,daw,0,dah

        # Visble Area intersecting the bounding box
        vxa,vya = max(px,px+xa-ox),   max(py,py+ya-oy)
        vxb,vyb = min(px+pw,vxa+xb-xa),min(py+ph,vya+yb-ya)
        vw,vh   = vxb-vxa+1, vyb-vya+1

        outData  = {
            'version':'1.1.0',
            'size':[vw,vh],
            'pos': (vxa,vya),
            'name':str(self.name()),
            'data':[], 'colors':[]}

        if palette:
            palette = outData['palette'] = []
            for row in colors:
                for c in row:
                    fg = f"{c.getHex(ttk.TTkK.Foreground)}" if c.foreground() else None
                    bg = f"{c.getHex(ttk.TTkK.Background)}" if c.background() else None
                    if (pc:=(fg,bg)) not in palette:
                        palette.append(pc)

        if full:
            wslice = slice(xa,xb+1)
            hslice = slice(ya,yb+1)
            outData['offset'] = (max(0,ox-xa),max(0,oy-ya))
        else:
            wslice = slice(ox+vxa-px,ox+vxa-px+vw)
            hslice = slice(oy+vya-py,oy+vya-py+vh)

        for row in data[hslice]:
            outData['data'].append(row[wslice])
        for row in colors[hslice]:
            outData['colors'].append([])
            for c in row[wslice]:
                fg = f"{c.getHex(ttk.TTkK.Foreground)}" if c.foreground() else None
                bg = f"{c.getHex(ttk.TTkK.Background)}" if c.background() else None
                if palette:
                    outData['colors'][-1].append(palette.index((fg,bg)))
                else:
                    outData['colors'][-1].append((fg,bg))

        return outData

    def _import_v1_1_0(self, dd):
        self._import_v1_0_0(dd)
        self._offset = dd.get('offset',(0,0))

    def _import_v1_0_0(self, dd):
        self._pos  = dd['pos']
        self._size = dd['size']
        self._name = dd['name']
        self._data = dd['data']
        def _getColor(cd):
            fg,bg = cd
            if fg and bg: return ttk.TTkColor.fg(fg)+ttk.TTkColor.bg(bg)
            elif fg:      return ttk.TTkColor.fg(fg)
            elif bg:      return ttk.TTkColor.bg(bg)
            else:         return ttk.TTkColor.RST
        if 'palette' in dd:
            palette = [_getColor(c) for c in  dd['palette']]
            self._colors = [[palette[c] for c in row] for row in dd['colors']]
        else:
            self._colors = [[_getColor(c) for c in row] for row in dd['colors']]

    def _import_v0_0_0(self, dd):
        # Legacy old import
        w = len(dd['data'][0]) + 10
        h = len(dd['data']) + 4
        x,y=5,2
        self.resize(w,h)
        self._pos = (0,0)
        for i,rd in enumerate(dd['data']):
            for ii,cd in enumerate(rd):
                self._data[i+y][ii+x] = cd
        for i,rd in enumerate(dd['colors']):
            for ii,cd in enumerate(rd):
                fg,bg = cd
                if fg and bg:
                    self._colors[i+y][ii+x] = ttk.TTkColor.fg(fg)+ttk.TTkColor.bg(bg)
                elif fg:
                    self._colors[i+y][ii+x] = ttk.TTkColor.fg(fg)
                elif bg:
                    self._colors[i+y][ii+x] = ttk.TTkColor.bg(bg)
                else:
                    self._colors[i+y][ii+x] = ttk.TTkColor.RST

    def importLayer(self, dd):
        self.clean()

        if 'version' in dd:
            ver = dd['version']
            if   ver == ('1.0.0'):
                self._import_v1_0_0(dd)
            elif ver == ('1.1.0'):
                self._import_v1_1_0(dd)
        else:
            self._import_v0_0_0(dd)
        self.changed.emit()

    def trim(self):
        pw,ph = self._size
        ox,oy = self._offset
        cl = CanvasLayer()
        if not (pw and ph) : return cl

        tmpd = [r[ox:ox+pw] for r in self._data[  oy:oy+ph]]
        tmpc = [r[ox:ox+pw] for r in self._colors[oy:oy+ph]]
        # Trim
        xa,xb,ya,yb = 0x10000,0,0x10000,0
        for y,(drow,crow) in enumerate(zip(tmpd,tmpc)):
            for x,(d,c) in enumerate(zip(drow,crow)):
                if d != ' ' or c.background():
                    xa,xb = min(x,xa),max(x,xb)
                    ya,yb = min(y,ya),max(y,yb)
        if (xa,xb,ya,yb) == (0x10000,0,0x10000,0):
            xa=xb=ya=yb=0

        cl._data   = [r[xa:xb+1] for r in tmpd[ya:yb+1]]
        cl._colors = [r[xa:xb+1] for r in tmpc[ya:yb+1]]

        w = xb-xa+1
        h = yb-ya+1

        cl._size = (w,h)
        cl._name = self._name
        return cl

    def importTTkString(self, txt:ttk.TTkString):
        tmpd = []
        tmpc = []
        for line in txt.split('\n'):
            d,c = line.getData()
            tmpd.append(list(d))
            tmpc.append(list(c))
        # Trim
        xa,xb,ya,yb = 0x10000,0,0x10000,0
        for y,(drow,crow) in enumerate(zip(tmpd,tmpc)):
            for x,(d,c) in enumerate(zip(drow,crow)):
                if d != ' ' or c.background():
                    xa,xb = min(x,xa),max(x,xb)
                    ya,yb = min(y,ya),max(y,yb)
        if (xa,xb,ya,yb) == (0x10000,0,0x10000,0):
            xa=xb=ya=yb=0

        self._data   = data   = [r[xa:xb+1] for r in tmpd[ya:yb+1]]
        self._colors = colors = [r[xa:xb+1] for r in tmpc[ya:yb+1]]

        w = xb-xa+1
        h = yb-ya+1

        for i,(d,c) in enumerate(zip(data,colors)):
            data[i]   = (d + [' ']*w)[:w]
            colors[i] = (c + [ttk.TTkColor.RST]*w)[:w]

        self._size = (w,h)
        self._name = "Pasted"

    def placeFill(self,geometry,tool,glyph:str,color:ttk.TTkColor,glyphEnabled=True,preview=False):
        ox,oy = self._offset
        w,h = self._size
        ax,ay,bx,by = geometry
        ax = max(0,min(w-1,ax))
        ay = max(0,min(h-1,ay))
        bx = max(0,min(w-1,bx))
        by = max(0,min(h-1,by))
        fax,fay = ox+min(ax,bx), oy+min(ay,by)
        fbx,fby = ox+max(ax,bx), oy+max(ay,by)

        if preview:
            data   = [_r.copy() for _r in self._data]
            colors = [_r.copy() for _r in self._colors]
            self._preview = {'data':data,'colors':colors}
        else:
            self._preview = None
            data   = self._data
            colors = self._colors

        if tool == ToolType.RECTFILL:
            for y in range(fay,fby+1):
                for x in range(fax,fbx+1):
                    self._placeGlyph(data,colors,x,y,glyph,color,glyphEnabled,preview)
        if tool == ToolType.RECTEMPTY:
            for x in range(fax,fbx+1):
                self._placeGlyph(data,colors,x,fay,glyph,color,glyphEnabled,preview)
                self._placeGlyph(data,colors,x,fby,glyph,color,glyphEnabled,preview)
            for y in range(fay,fby+1):
                self._placeGlyph(data,colors,fax,y,glyph,color,glyphEnabled,preview)
                self._placeGlyph(data,colors,fbx,y,glyph,color,glyphEnabled,preview)
        self.changed.emit()
        return True

    def placeGlyph(self,x,y,glyph:str,color:ttk.TTkColor,glyphEnabled=True,preview=False):
        if preview:
            data   = [_r.copy() for _r in self._data]
            colors = [_r.copy() for _r in self._colors]
            self._preview = {'data':data,'colors':colors}
        else:
            self._preview = None
            data   = self._data
            colors = self._colors

        self.changed.emit()
        return self._placeGlyph(data,colors,x,y,glyph,color,glyphEnabled,preview)

    def _placeGlyph(self,data,colors,x,y,glyph:str,color:ttk.TTkColor,glyphEnabled=True,preview=False):
        ox,oy = self._offset
        w,h = self._size

        if 0<=x<w and 0<=y<h:
            if glyphEnabled:
                color = color if glyph != ' ' else color.background()
                color = color if color else ttk.TTkColor.RST
                data[  oy+y][ox+x] = glyph
                colors[oy+y][ox+x] = color
            else:
                glyph = data[  oy+y][ox+x]
                ofg = colors[oy+y][ox+x].foreground()
                obg = colors[oy+y][ox+x].background()
                nfg = color.foreground()
                nbg = color.background()
                if glyph==' ':
                    if obg:
                        colors[oy+y][ox+x] = nbg if nbg else ttk.TTkColor.RST
                else:
                    fg = nfg if nfg         else ofg if ofg else ttk.TTkColor.RST
                    bg = nbg if nbg and obg else obg if obg else fg
                    color = fg+bg
                    color = color if color else ttk.TTkColor.RST
                    colors[oy+y][ox+x] = color
            return True
        return False

    def placeArea(self,x,y,area,transparent=False,preview=False):
        ox,oy = self._offset
        w,h = self._size

        dw,dh = area.size()
        darea = area._data
        carea = area._colors
        x-=dw//2
        y-=dh//2

        if preview:
            data   = [_r.copy() for _r in self._data]
            colors = [_r.copy() for _r in self._colors]
            self._preview = {'data':data,'colors':colors}
        else:
            self._preview = None
            data   = self._data
            colors = self._colors

        for _y,(darow,carow) in enumerate(zip(darea,carea),oy+y):
            for _x,(da,ca)   in enumerate(zip(darow,carow),ox+x):
                if 0<=_x<w and 0<=_y<h and ( da!=' ' or ca.background()):
                    if not transparent or (da==' ' and ca._bg):
                        data[_y][_x]   = da
                        colors[_y][_x] = ca
                    elif da!=' ':
                        data[_y][_x]   = da
                        cc = colors[_y][_x]
                        newC = ca.copy()
                        newC._bg = ca._bg if ca._bg else cc._bg
                        colors[_y][_x] = newC
        self.changed.emit()

    def drawInCanvas(self, pos, canvas:ttk.TTkCanvas):
        if not self._visible: return
        px,py = pos
        pw,ph = self._size
        cw,ch = canvas.size()
        if px+pw<0 or py+ph<0:return
        if px>=cw or py>=ch:return
        # Data Offset
        ox,oy = self._offset
        # x,y position in the Canvas
        cx = max(0,px)
        cy = max(0,py)
        # x,y position in the Layer
        lx,ly = (cx-px),(cy-py)
        # Area to be copyed
        dw = min(cw-cx,pw-lx)
        dh = min(ch-cy,ph-ly)

        if _p := self._preview:
            data   = _p['data']
            colors = _p['colors']
        else:
            data   = self._data
            colors = self._colors
        for y in range(cy,cy+dh):
            for x in range(cx,cx+dw):
                gl = data[  oy+y+ly-cy][ox+x+lx-cx]
                c  = colors[oy+y+ly-cy][ox+x+lx-cx]
                if gl==' ' and c._bg:
                    canvas._data[y][x]   = gl
                    canvas._colors[y][x] = c
                elif gl!=' ':
                    canvas._data[y][x]   = gl
                    cc = canvas._colors[y][x]
                    newC = c.copy()
                    newC._bg = c._bg if c._bg else cc._bg
                    canvas._colors[y][x] = newC



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['About']

import TermTk as ttk

_DPT_image = ttk.TTkUtil.base64_deflate_2_obj(
    "eJydWDtuJDcQdeBIuaNNdAKhi78mwTs4M2BgYwNOlPkAo10FC3sAK7C0M1jJAjacC2y+R9FJXFVksYtU9wiwRA40XcXm5716VdTux/tffvqBf379Gf949366vsSfd+9t" +
    "zCbPU44hg79+eXiqj8DNGcKcZ5cdfTcxW5cDeuxwcGYfGjPNOeHD2/JGGTxhn/Ev76//+P03tn3/Vo3G4bBk2YgDW6tretl9bZ+Xq0M+V+eri2Eb0WcwU3Z1QWUp3mWI" +
    "CQ2gnpopZWPQPcWyPVxswjli3UrZHwQcO+NjN+Hzo7wRsNuQo1MvnOhXfbe2tFdHY9yE73N1K5+wkfnl4XTeYWP8iW2rpg9nhn18a9iq6Z6X89AjXVAmxJc5x591X243" +
    "NOJCzwkwZ8dn+VmehEQMDQTrc1sZQjpF7L4AiKvsETBIV4TAQ7ED+maa90Hs3hJhomt2/NDjPX5PZLDFY8LuOwdPzOBeODSuwOMQbHOdYaaXoP1fZYeZKFZnMMhVPKM5" +
    "DrMgBedZXDAKOAz37VQ9L81V2pU1PDQSYwxz01u3ntheUL1DVCk2FeqjGScrn5texQV9d8LXNa8/z43HkY8SLKuEQZ9jUYq63HVm9XvpXnH/18LP1SmeeJZD4+QQ3I2Q" +
    "FrUxEWUKJggj2xs5EcfEzQmE+NwSfx8Vu1CVsBX2Zsvk3Dcz4CwoWpU69Cp2uFMOSM4ECy0cRkLV50ZwsNTd4kT0SqMT4P5BOQlfdoqDIWIAVhdH6aGki0NDExcAtM26" +
    "n0mWK7nC1y1+6XUJ84qPSuyuOo6M1k1joentOZdty4FzypmRonpvvhY8JylUnbqmTeO41M74YdDhznjafuem5aaILWbVDZfv3ySITyuKjFzPopP0q7gBlDaReiKzIkCN" +
    "OwClV4HCOGD7ocWS5z6JDI+ZlHI/dZk/Je7dHJTjkX0mLDqKEJmksraxwF0ChsobGqVkEaIhS53HEItdIWxdacptPJ0c7qYUTKwpjRuOtKFwY9e1evjnPff8eejpphyf" +
    "2H4Ssd0pJEyRMjyZAebFwEFUSrhVM4bBR9SxQeq1xx0vsQjlF5bs46C4/WxFsp/XPcrInmwVfqlDHT+oh15td+1Fnns1YzmKTTtQ7RAFUYfFwCzJcakNvWvcs1xSdDOQ" +
    "jFtxMJOpK2i1IOZl8G2BVDBPXRWB35CVYs7BsnW/WCn/i2qWFZk+fWcS3zJlzV1KlZfUdbtSSo8F2NqIVkbgzg3MgjEidxpwVQ5IkBtSjOOWwyvpU7ZPNPLEU/xdisrt" +
    "iSRoLvqiM+ZifmyhGbkLEbCySH3GxiKO+iTpiSSki33EyTVxoNRqVEIngA14qQMhzZ1EoXGZnGo+6GhmKYRpX0HRoGStjiWks65NZ2O3APBmmSGVQnLfWIL+qT26XoGc" +
    "ssDu69p9QjnU1LIx/A0rKwuhVdJ94tIKL2miNsstTUy9Ph6Ktmkm9J6H/2U6CtGfFY+WzBa5Cy0SNdDxT/E9zwpbA3LOlXn4aOILb8tvRF/bRfnEtX7DFjrwkXfUvFPg" +
    "T0mNp+QXuOJb3jBBR5+pVIDL+LmbH5NdlASdk+ctHht5SjVnNHtGhCWmr8Yrp7LfShl/u0qwFZcVj5KfXtcgqipQB7nY+dpEkdZsVlcpCCEhaSS5V+F3umqv0HkvyQZX" +
    "5bp/iXDrcQCNpA/U6GraSiXjgk4XVHH0UANXKR2SeheQRiaBC4s5jjIwbQFJt89O3vckw/XPK8nrG9/LVw2KhkBjZJ1Gp98YeL8Y0+tt61NJQR18CNQ6hoNzOkzxOfV2" +
    "skC5nGLTO3X8g4zL6WqVh6TjNJjx8I1aQ/Dd4XNlvHr4JEeD5L36J4qyl4vMbqz0t5yuupeuRs2TunhK62B60nsEBQRLe2cNs6a/1WV44maCZreOD/7PGZgFSUt3Qq8D" +
    "lSEyctFkgU3qtlCgVwE4h2FwTIMEHDqA7HVH5MuVkxqq6kCNk/Fu8G1H/s/Vf0gffCQ=")

class About(ttk.TTkWindow):
    def __init__(self, **kwargs):
        image = ttk.TTkLabel(text=_DPT_image)
        super().__init__(**kwargs)
        self.setTitle('About the Dumb Paint Tool...')
        self.resize(65+2,12+4)
        self.layout().addWidgets([
            ttk.TTkLabel(text=_DPT_image),
            ttk.TTkLabel(pos=(38,7), text=ttk.TTkString(f"  Version: {ttk.TTkCfg.version}", ttk.TTkColor.fg('#AAAAFF'))),
            ttk.TTkLabel(pos=(38,9), text=ttk.TTkString("Powered By, Eugenio Parodi",ttk.TTkColor.fg("#FF00FF"))),
            ttk.TTkLabel(pos=(13,11),text=ttk.TTkString("https://github.com/ceccopierangiolieugenio/pyTermTk", ttk.TTkColor.fg('#44FFFF')))
            ])



# MIT License
#
# Copyright (c) 2024 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['PaintTemplate']

import os, json

import TermTk as ttk

from .paintarea    import PaintArea, PaintScrollArea
from .toolspanel   import ToolsPanel
from .canvaslayer  import CanvasLayer
from .painttoolkit import PaintToolKit
from .textarea     import TextArea
from .layers       import Layers,LayerData
from .about        import About
from .const        import ToolType
from .filters      import HueChromaLightness,BrightnessContrast
from .importimage  import ImportImage
from .exportarea   import ExportArea


# Layout:
#
#  Palette                           Brushes
#                Drawing Area        (Chars/Glyphs)
#  Tools
#
#                                    Layouts
#                 Export
#
class PaintTemplate(ttk.TTkAppTemplate):
    __slots__ = ('_parea','_layers')
    def __init__(self, fileName=None, border=False, **kwargs):
        super().__init__(border, **kwargs)
        self._parea  = parea = PaintArea()
        self._layers = layers = Layers()
        ptoolkit = PaintToolKit()
        tarea    = TextArea()
        expArea  = ExportArea(parea)

        toolsPanel = ToolsPanel()

        rightPanel = ttk.TTkSplitter(orientation=ttk.TTkK.VERTICAL)
        rightPanel.addWidget(tarea)
        # rightPanel.addItem(expArea, title="Export")
        # rightPanel.setSizes([None,5])
        rightPanel.addItem(layers, title='Layers')
        rightPanel.setSizes([None,9])

        self.setItem(expArea, self.BOTTOM, title="Export")

        self.setItem(toolsPanel     , self.LEFT,  size=16*2)
        self.setWidget(PaintScrollArea(parea) , self.MAIN)
        self.setWidget(ptoolkit      , self.TOP,   fixed=True)
        self.setItem(rightPanel    , self.RIGHT, size=40)

        self.setMenuBar(appMenuBar:=ttk.TTkMenuBarLayout(), self.TOP)
        fileMenu      = appMenuBar.addMenu("&File")
        fileMenu.addMenu("&New"       ).menuButtonClicked.connect(self._new)
        fileMenu.addMenu("&Open"      ).menuButtonClicked.connect(self._open)
        fileMenu.addMenu("&Save"      ).menuButtonClicked.connect(self._save)
        fileMenu.addMenu("Save &As...").menuButtonClicked.connect(self._saveAs)
        fileMenu.addSpacer()
        fileMenu.addMenu("&Import").menuButtonClicked.connect(self.importDictWin)
        fileMenu.addMenu("Import Image").menuButtonClicked.connect(self._openImage)
        menuExport = fileMenu.addMenu("&Export")
        fileMenu.addSpacer()
        fileMenu.addMenu("Load Palette").setEnabled(False)
        fileMenu.addMenu("Save Palette").setEnabled(False)
        fileMenu.addSpacer()
        buttonExit    = fileMenu.addMenu("E&xit")
        buttonExit.menuButtonClicked.connect(ttk.TTkHelper.quit)

        menuExport.addMenu("Ascii/&Txt").menuButtonClicked.connect(self._saveAsAscii)
        menuExport.addMenu("&Ansi").menuButtonClicked.connect(self._saveAsAnsi)
        menuExport.addMenu("&Python").setEnabled(False)
        menuExport.addMenu("&Bash").setEnabled(False)

        colorMenu      = appMenuBar.addMenu("&Color")
        colorMenu.addMenu("&Hue Chroma Lightness").menuButtonClicked.connect(self._hueChromaLightness)
        colorMenu.addMenu("&Brightness Contrast").menuButtonClicked.connect(self._brightnessContrast)

        # extraMenu = appMenuBar.addMenu("E&xtra")
        # extraMenu.addMenu("Scratchpad").menuButtonClicked.connect(self.scratchpad)
        # extraMenu.addSpacer()

        def _showAbout(btn):
            ttk.TTkHelper.overlay(None, About(), 30,10)
        def _showAboutTTk(btn):
            ttk.TTkHelper.overlay(None, ttk.TTkAbout(), 30,10)

        helpMenu = appMenuBar.addMenu("&Help", alignment=ttk.TTkK.RIGHT_ALIGN)
        helpMenu.addMenu("About ...").menuButtonClicked.connect(_showAboutTTk)
        helpMenu.addMenu("About DPT").menuButtonClicked.connect(_showAbout)

        # Paint Area Transparency color
        ptoolkit.updatedTrans.connect(self._parea.setTrans)

        @ttk.pyTTkSlot(CanvasLayer)
        def _canvasLayerSelected(l:CanvasLayer):
            layers.selectLayerByData(l)

        @ttk.pyTTkSlot(LayerData)
        def _layerDataSelected(l:LayerData):
            parea.setCurrentLayer(l.data())

        parea.layerSelected.connect(ptoolkit.updateLayer)
        parea.layerSelected.connect(_canvasLayerSelected)
        parea.layerAdded.connect(self._canvasLayerAdded)

        layers.layerAdded.connect(self._layerAdded)
        layers.layerSelected.connect(_layerDataSelected)
        layers.layersOrderChanged.connect(self._layersOrderChanged)
        layers.addLayer(name="Background")
        if fileName:
            self._openFile(fileName)

        ttk.ttkConnectDragOpen(ttk.TTkEncoding.APPLICATION_JSON, self._openDragData)
        ttk.ttkConnectDragOpen(ttk.TTkEncoding.IMAGE, self._openImageData)

        # Debug import image
        # from PIL import Image
        # pilImage = Image.open("experiments/Peppered/euDock-purple.png")
        # newWindow = ImportImage(pilImage,minSize=(60,30))
        # newWindow.exportedImage.connect(parea.pasteEvent)
        # ttk.TTkHelper.overlay(None, newWindow, 10, 4, modal=True)

    @ttk.pyTTkSlot()
    def _new(self):
        self._parea.clear()
        self._layers.clear()
        self._layers.addLayer(name="Background")

    @ttk.pyTTkSlot()
    def _open(self):
        ttk.ttkCrossOpen(
                path='.',
                encoding=ttk.TTkEncoding.APPLICATION_JSON,
                filter="DumbPaintTool Files (*.DPT.json);;Json Files (*.json);;All Files (*)",
                cb=self._openDragData)

    @ttk.pyTTkSlot()
    def _openImage(self):
        ttk.ttkCrossOpen(
                path='.',
                encoding=ttk.TTkEncoding.IMAGE,
                filter="Images (*.png *.jpg *.gif *.ico);;All Files (*)",
                cb=self._openImageData)

    @ttk.pyTTkSlot()
    def _save(self):
        doc = self._parea.exportDocument()
        ttk.ttkCrossSave('untitled.DPT.json', json.dumps(doc, indent=1), ttk.TTkEncoding.APPLICATION_JSON)

    @ttk.pyTTkSlot()
    def _saveAs(self):
        doc = self._parea.exportDocument()
        ttk.ttkCrossSaveAs('untitled.DPT.json', json.dumps(doc, indent=1), ttk.TTkEncoding.APPLICATION_JSON,
                           filter="DumbPaintTool Files (*.DPT.json);;Json Files (*.json);;All Files (*)")

    @ttk.pyTTkSlot()
    def _saveAsAnsi(self):
        image = self._parea.exportImage()
        text = ttk.TTkString(image)
        ttk.ttkCrossSaveAs('untitled.DPT.Ansi.txt', text.toAnsi(), ttk.TTkEncoding.TEXT_PLAIN_UTF8,
                           filter="Ansi text Files (*.Ansi.txt);;Text Files (*.txt);;All Files (*)")

    @ttk.pyTTkSlot()
    def _saveAsAscii(self):
        image = self._parea.exportImage()
        text = ttk.TTkString(image)
        ttk.ttkCrossSaveAs('untitled.DPT.ASCII.txt', text.toAscii(), ttk.TTkEncoding.TEXT_PLAIN_UTF8,
                           filter="ASCII Text Files (*.ASCII.txt);;Text Files (*.txt);;All Files (*)")

    @ttk.pyTTkSlot(dict)
    def _openImageData(self, data):
        newWindow = ImportImage(data['data'])
        newWindow.exportedImage.connect(self._parea.pasteEvent)
        ttk.TTkHelper.overlay(None, newWindow, 10, 4, modal=True)

    @ttk.pyTTkSlot(dict)
    def _openDragData(self, data):
        dd = json.loads(data['data'])
        if 'layers' in dd:
            self.importDocument(dd)
        else:
            self._layers.addLayer(name="Import")
            self._parea.importLayer(dd)

    def _openFile(self, fileName):
        ttk.TTkLog.info(f"Open: {fileName}")

        with open(fileName) as fp:
            # dd = json.load(fp)
            # text = fp.read()
            # dd = eval(text)
            dd = json.load(fp)
            if 'layers' in dd:
                self.importDocument(dd)
            else:
                self._layers.addLayer(name="Import")
                self._parea.importLayer(dd)

    # Connect and handle Layers event
    @ttk.pyTTkSlot(LayerData)
    def _layerAdded(self, l:LayerData):
        self._parea.layerAdded.disconnect(self._canvasLayerAdded)
        nl = self._parea.newLayer()
        self._parea.layerAdded.connect(self._canvasLayerAdded)
        nl.setName(l.name())
        l.setData(nl)
        l.nameChanged.connect(nl.setName)
        l.visibilityToggled.connect(nl.setVisible)
        l.visibilityToggled.connect(self._parea.update)

    @ttk.pyTTkSlot(CanvasLayer)
    def _canvasLayerAdded(self, l:CanvasLayer=None):
        self._layers.clear()
        self._layers.layerAdded.disconnect(self._layerAdded)
        for l in self._parea.canvasLayers():
            ld = self._layers.addLayer(name=l.name(),data=l)
            ld.nameChanged.connect(l.setName)
            ld.visibilityToggled.connect(l.setVisible)
            ld.visibilityToggled.connect(self._parea.update)
        self._layers.layerAdded.connect(self._layerAdded)

    @ttk.pyTTkSlot(list[LayerData])
    def _layersOrderChanged(self, layers:list[LayerData]):
        self._parea._canvasLayers = [ld.data() for ld in reversed(layers)]
        self._parea.update()

    @ttk.pyTTkSlot()
    def _hueChromaLightness(self):
        newWindow = HueChromaLightness(self._parea._currentLayer)
        ttk.TTkHelper.overlay(None, newWindow, 10, 4, modal=True)

    @ttk.pyTTkSlot()
    def _brightnessContrast(self):
        newWindow = BrightnessContrast(self._parea._currentLayer)
        ttk.TTkHelper.overlay(None, newWindow, 10, 4, modal=True)

    def importDocument(self, dd):
        self._parea.layerAdded.disconnect(self._canvasLayerAdded)
        self._parea.importDocument(dd)
        self._canvasLayerAdded()
        self._parea.layerAdded.connect(self._canvasLayerAdded)

    @ttk.pyTTkSlot()
    def importDictWin(self):
        newWindow = ttk.TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/quickImport.tui.json"))
        te = newWindow.getWidgetByName("TextEdit")

        @ttk.pyTTkSlot()
        def _importDict(te=te):
            def _probeCompressedText(_text):
                import re
                ret = ""
                for _t in _text.split('\n'):
                    if m := re.match(r'^ *["\']([A-Za-z0-9+/]+[=]{0,2})["\' +]*$',_t):
                        ret += m.group(1)
                    elif not re.match(r'^ *$',_t): # exclude empty lines
                        return ""
                return ret
            text = te.toPlainText()
            if compressed := _probeCompressedText(text):
                dd = ttk.TTkUtil.base64_deflate_2_obj(compressed)
            else:
                try:
                    dd = eval(text)
                except Exception as e:
                    ttk.TTkLog.error(str(e))
                    messageBox = ttk.TTkMessageBox(text= str(e),icon=ttk.TTkMessageBox.Icon.Warning)
                    ttk.TTkHelper.overlay(None, messageBox, 5, 5, True)
                    return

            if type(dd) is not dict:
                messageBox = ttk.TTkMessageBox(text= f"Input is {type(dd)}\nImport data must be a dict or \ncompressed String definition",icon=ttk.TTkMessageBox.Icon.Warning)
                ttk.TTkHelper.overlay(None, messageBox, 5, 5, True)
                return

            if 'layers' in dd:
                self.importDocument(dd)
            else:
                self._layers.addLayer(name="Import")
                self._parea.importLayer(dd)

            newWindow.close()

        newWindow.getWidgetByName("BtnDict"   ).clicked.connect(_importDict)
        ttk.TTkHelper.overlay(None, newWindow, 10, 4, modal=True)

